#!/usr/bin/env python3
# ruff: noqa: N999
import argparse
import base64
import contextlib
import csv
import dataclasses
import datetime
import errno
import getpass
import io
import logging
import math
import os
import re
import sys
import tarfile
import tempfile
import time
import urllib.request
import urllib.response
from collections.abc import Generator
from typing import IO, Any, Callable, Optional, Union, cast

import numpy as np

__all__ = []


def export(obj):
    name = obj if isinstance(obj, str) else obj.__name__
    __all__.append(name)
    return obj


api_url = "https://hsc-release.mtk.nao.ac.jp/das_cutout/pdr3"

available_reruns = [
    "pdr3_dud_rev",
    "pdr3_dud",
    "pdr3_wide",
]

default_rerun = "pdr3_wide"

available_types = [
    "coadd",
    "coadd/bg",
    "warp",
]

default_type = "coadd"

default_get_image = True
default_get_mask = False
default_get_variance = False

default_name = "{lineno}_{type}_{ra:.5f}_{dec:+.5f}_{tract}_{filter}"

default_max_connections = 4

export("ANYTRACT")
ANYTRACT = -1
export("NODIM")
NODIM = (-1, -1)
export("ALLFILTERS")
ALLFILTERS = "all"

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        fromfile_prefix_chars="@",
        description="""
            Download FITS cutouts from the website of HSC data release.
        """,
    )
    parser.add_argument(
        "--ra",
        metavar="DEGREES",
        type=parse_longitude,
        help="""
        R.A.2000.
    """,
    )
    parser.add_argument(
        "--dec",
        metavar="DEGREES",
        type=parse_latitude,
        help="""
        Dec.2000.
    """,
    )
    parser.add_argument(
        "--sw",
        metavar="DEGREES",
        type=parse_degree,
        help="""
        Semi-width in R.A. direction.
    """,
    )
    parser.add_argument(
        "--sh",
        metavar="DEGREES",
        type=parse_degree,
        help="""
        Semi-height in Dec. direction.
    """,
    )
    parser.add_argument(
        "--filter",
        type=parse_filter_opt,
        help="""
        Filter name.
    """,
    )
    parser.add_argument(
        "--rerun",
        choices=available_reruns,
        default=default_rerun,
        help="""
        Rerun name.
    """,
    )
    parser.add_argument(
        "--tract",
        type=parse_tract_opt,
        help="""
        Tract number.
    """,
    )
    parser.add_argument(
        "--image",
        metavar="BOOL",
        type=parse_bool,
        default=default_get_image,
        help=f"""
        Get the image layer. (Default: {default_get_image})
    """,
    )
    parser.add_argument(
        "--mask",
        metavar="BOOL",
        type=parse_bool,
        default=default_get_mask,
        help=f"""
        Get the mask layer. (Default: {default_get_mask})
    """,
    )
    parser.add_argument(
        "--variance",
        metavar="BOOL",
        type=parse_bool,
        default=default_get_variance,
        help=f"""
        Get the variance layer. (Default: {default_get_variance})
    """,
    )
    parser.add_argument(
        "--type",
        choices=available_types,
        default=default_type,
        help="""
        Data type.
    """,
    )
    parser.add_argument(
        "--name",
        type=str,
        default=default_name,
        help=f"""
        Output name. (python's format string; default: "{default_name}")
    """,
    )
    parser.add_argument(
        "--list",
        metavar="PATH",
        type=str,
        help="""
        Path to a coordinate list.
        If this list is given, the other command-line arguments are optional.
        Missing fields in the list will default
        to the values given in the command-line.
    """,
    )
    parser.add_argument(
        "--listtype",
        choices=["auto", "txt", "csv"],
        default="auto",
        help="""
        How to interpret the argument of --list.
        "auto" (default): Follow the extension of the file name. /
        "txt": Fields are separated by one or more spaces. /
        "csv": Comma-separated volume.
    """,
    )
    parser.add_argument(
        "--user",
        type=str,
        help="""
        User account.
    """,
    )
    parser.add_argument(
        "--password",
        type=str,
        help="""
        Password.
        If you specify --password, your password is disclosed to everybody
        on the computer you use.
        Use of --password-env is recommended instead,
        especially when this script is run on a shared-use computer.
    """,
    )
    parser.add_argument(
        "--password-env",
        metavar="ENV",
        type=str,
        default="HSC_SSP_CAS_PASSWORD",
        help="""
        Name of the environment variable from which to read password.
        Use `read -s HSC_SSP_CAS_PASSWORD` to put your password into
        $HSC_SSP_CAS_PASSWORD.
    """,
    )
    parser.add_argument(
        "--semaphore",
        metavar="PATH",
        type=str,
        default="",
        help=f"""
        Path to the named semaphore (This is not a Posix semaphore.)
        The default name is `/tmp/$(id -u -n)-downloadCutout`.
        This path must be in NFS (or any other shared filesystem)
        if you distribute the processes over a network.
        If you specify this option, `--max-connections` is {default_max_connections} by default.
    """,
    )
    parser.add_argument(
        "--max-connections",
        metavar="NUM",
        type=int,
        default=0,
        help="""
        Maximum number of connections in parallel.
        This script itself won't make parallel connections,
        but _you_ should launch this script in parallel.
        The launched processes will communicate with each other to limit
        the number of connections in parallel.
        By default `--max-connections=0`, which means limitless.
    """,
    )

    args = parser.parse_args()

    rect = Rect.create(
        rerun=args.rerun,
        type=args.type,
        filter=args.filter,
        tract=args.tract,
        ra=args.ra,
        dec=args.dec,
        sw=args.sw,
        sh=args.sh,
        image=args.image,
        mask=args.mask,
        variance=args.variance,
        name=args.name,
    )

    if not args.password:
        args.password = os.environ.get(args.password_env)

    if args.list:
        with open_inputfile(sys.stdin if args.list == "-" else args.list) as f:
            rects = read_rects(f, default=rect, type=args.listtype)
    else:
        if not rect.iscomplete():
            raise RuntimeError("Specify either (--ra --dec --sw --sh) or --list.")
        rects = [rect]

    if args.semaphore and not args.max_connections:
        args.max_connections = default_max_connections
    if args.max_connections:
        set_max_connections(args.max_connections, args.semaphore)

    download(rects, user=args.user, password=args.password, onmemory=False)


@export
@dataclasses.dataclass(order=True)
class Rect:
    """
    Rectangle to cut out of the sky.

    The constructor is not intended to be called by usual users.
    Use Rect.create() instead.

    Parameters
    ----------
    rerun
        Rerun name.
    type
        "coadd" (coadd),
        "coadd/bg" (coadd with background), or
        "warp" (warp).
    filter
        Filter name.
        This member can be `ALLFILTERS`.
    tract
        Tract number.
        This member can be `ANYTRACT`.
    ra
        R.A.2000, in degrees.
    dec
        Dec.2000, in degrees.
    sw
        Semi-width, in degrees.
    sh
        Semi-height, in degrees.
    image
        Whether to get the image layer.
    mask
        Whether to get the mask layer.
    variance
        Whether to get the variance layer.
    name
        File name format (without extension ".fits")
    lineno
        Line number in a list file.
    dim
        Dimensions of downloaded file
    """

    rerun: str = default_rerun
    type: str = default_type
    tract: int = ANYTRACT
    filter: Union[str, list[str]] = ALLFILTERS
    ra: float = math.nan
    dec: float = math.nan
    sw: float = math.nan
    sh: float = math.nan
    image: bool = default_get_image
    mask: bool = default_get_mask
    variance: bool = default_get_variance
    name: str = default_name
    lineno: int = 0
    dim: tuple[int, int] = NODIM

    @staticmethod
    def create(
        rerun: Union[str, None] = None,
        type: Union[str, None] = None,
        tract: Union[str, int, None] = None,
        filter: Union[Union[str, list[str]], None] = None,
        ra: Union[str, float, None] = None,
        dec: Union[str, float, None] = None,
        sw: Union[str, float, None] = None,
        sh: Union[str, float, None] = None,
        image: Union[str, bool, None] = None,
        mask: Union[str, bool, None] = None,
        variance: Union[str, bool, None] = None,
        name: Union[str, None] = None,
        lineno: Union[int, None] = None,
        dim: Union[tuple[int, int], None] = None,
        default: Union["Rect", None] = None,
    ) -> "Rect":
        """
        Create a Rect object.

        If any parameter is omitted,
        it defaults to the corresponding field of the `default` argument.

        Parameters
        ----------
        rerun
            Rerun name.
        type
            "coadd" (coadd),
            "coadd/bg" (coadd with background), or
            "warp" (warp).
        filter
            Filter name.
            This member can be `ALLFILTERS`.
        tract
            Tract number.
            This member can be `ANYTRACT`.
        ra
            R.A.2000, in degrees.
            This argument can be a string like "12:34:56.789" (hours),
            "12h34m56.789s", "1.2345rad" (radians), etc.
        dec
            Dec.2000, in degrees.
            This argument can be a string like "-1:23:45.678" (degrees),
            "-1d23m45.678s", "1.2345rad" (radians), etc.
        sw
            Semi-width, in degrees.
            This argument can be a string like "1.2arcmin" etc.
        sh
            Semi-height, in degrees.
            This argument can be a string like "1.2arcmin" etc.
        image
            Whether to get the image layer.
        mask
            Whether to get the mask layer.
        variance
            Whether to get the variance layer.
        name
            File name format (without extension ".fits")
        lineno
            Line number in a list file.
        dim
            Dimensions of the image in pixels.
        default
            Default value.

        Returns
        -------
        rect
            Created `Rect` object.
        """
        rect = Rect() if default is None else Rect(*dataclasses.astuple(default))

        if rerun is not None:
            rect.rerun = parse_rerun(rerun)
        if type is not None:
            rect.type = parse_type(type)
        if filter is not None:
            rect.filter = parse_filter_opt(filter)
        if tract is not None:
            rect.tract = parse_tract_opt(tract)
        if ra is not None:
            rect.ra = parse_longitude(ra)
        if dec is not None:
            rect.dec = parse_latitude(dec)
        if sw is not None:
            rect.sw = parse_degree(sw)
        if sh is not None:
            rect.sh = parse_degree(sh)
        if image is not None:
            rect.image = parse_bool(image)
        if mask is not None:
            rect.mask = parse_bool(mask)
        if variance is not None:
            rect.variance = parse_bool(variance)
        if name is not None:
            rect.name = str(name)
        if lineno is not None:
            rect.lineno = int(lineno)
        if dim is not None:
            rect.dim = dim

        return rect

    def iscomplete(self) -> bool:
        """
        Whether or not `self` is complete.

        If, for example, user creates a `Rect` object:
            rect = Rect.create(sw=25, sh=25)
        then `rect` does not have valid values of `ra` and `dec`.
        In such a case, this function returns False.

        Returns
        -------
        iscomplete
            True if `self` is complete
        """
        return self.ra == self.ra and self.dec == self.dec and self.sw == self.sw and self.sh == self.sh

    def explode(self) -> list["Rect"]:
        """
        Make copies of `self` with more specific values.

        Returns
        -------
        rects
            List of `Rect` objects, each being more specific than `self`.
        """
        if isinstance(self.filter, list):
            return [Rect.create(filter=f, default=self) for f in self.filter]
        elif self.filter == ALLFILTERS:
            return [Rect.create(filter=f, default=self) for f in _all_filters]
        else:
            return [Rect.create(default=self)]

    # Static field list used by __eq__ and __hash__
    immutable_fields = ["ra", "dec", "sw", "sh", "filter", "type", "rerun", "image", "variance", "mask"]

    def __eq__(self, obj) -> bool:
        """Define equality on Rects by sky location, size, filter, type, rerun, and image/mask/variance state.
        This allows rects can be used as keys in dictionaries while ignoring transient fields such as lineno,
        or fields that may be incorrect/changed during download process like tract or name.

        This is a compromise between
        1) Dataclass's unsafe_hash=True which would hash all fields
        and
        2) Making the dataclass frozen which would affect some of the mutability used to alter
          lineno, tract, and name

        Note that this makes equality on Rects means "the cutout API should return the same data",
        rather than "Literally all data members the same"

        Parameters
        ----------
        obj : Rect
            The rect to compare to self

        Returns
        -------
        bool
            True if the Rect's are equal
        """
        return all([self.__dict__[field] == obj.__dict__[field] for field in Rect.immutable_fields])

    def __hash__(self):
        """Define a hash function on Rects. Outside of hash collisions, this function attempts to have the
        same semantics as Rect.__eq__(). Look at Rect.__eq__() for further details.
        """
        return hash(tuple([self.__dict__[field] for field in Rect.immutable_fields]))


@export
def read_rects(
    file: Union[str, IO], default: Optional[Rect] = None, type: Optional[str] = None
) -> list[Rect]:
    """
    Read a file to get a list of `Rect` objects.

    Parameters
    ----------
    file
        A file path or a file object.
    default
        Default values.
        Fields that cannot be obtained from the file
        defaults to the corresponding fields of this object.
    type
        File type. One of "auto" (=None), "txt", "csv".
        By default, the file type is guessed
        from the extension part of the file name.

    Returns
    -------
    rects
        List of `Rect` objects.
    """
    if (not type) or type == "auto":
        isfileobj = hasattr(file, "read")
        name = str(getattr(file, "name", "(file name not available)") if isfileobj else file)
        _, ext = os.path.splitext(name)
        type = ext.lstrip(".") or "txt"

    retval: list[Rect] = []

    if type == "txt":
        retval = read_rects_from_txt(file, default=default)
    elif type == "csv":
        retval = read_rects_from_csv(file, default=default)
    else:
        raise ValueError(f"Invalid file type: {type}")

    return retval


@export
def read_rects_from_txt(file, default=None) -> list[Rect]:
    """
    Read a space-separated volume to get a list of `Rect` objects.
    The first line must contain column names.

    Parameters
    ----------
    file
        A file path or a file object.
    default
        Default values.
        Fields that cannot be obtained from the file
        defaults to the corresponding fields of this object.

    Returns
    -------
    rects
        List of `Rect` objects.
    """
    allfields = set(field.name for field in dataclasses.fields(Rect))
    rects: list[Rect] = []
    with open_inputfile(file) as f:
        f_io = io.TextIOWrapper(f, encoding="utf-8")

        fieldnames = re.sub(r"^#\??\s*", "", f_io.readline().strip().lower()).split()
        validfields = [(i, str(field)) for i, field in enumerate(fieldnames) if field in allfields]
        if not validfields:
            raise RuntimeError("No column has a valid name in the list.")

        for lineno, line in enumerate(f, start=2):
            row = line.strip().split()
            if len(row) != len(fieldnames):
                raise RuntimeError(
                    f"line {lineno}: number of fields ({len(row)}) "
                    f"does not agree with what expected ({len(fieldnames)})"
                )
            args: dict[str, Any] = {"lineno": lineno}
            args.update((field, row[i]) for i, field in validfields)
            rects.append(Rect.create(default=default, **args))

    return rects


@export
def read_rects_from_csv(file, default=None) -> list[Rect]:
    """
    Read a comma-separated volume to get a list of `Rect` objects.
    The first line must contain column names.

    Parameters
    ----------
    file
        A file path or a file object.
    default
        Default values.
        Fields that cannot be obtained from the file
        defaults to the corresponding fields of this object.

    Returns
    -------
    rects
        List of `Rect` objects.
    """
    allfields = set(field.name for field in dataclasses.fields(Rect))

    rects: list[Rect] = []
    with open_inputfile(file) as f:
        reader = csv.reader(io.TextIOWrapper(f, encoding="utf-8", newline=""))

        fieldnames = next(reader)
        if len(fieldnames) > 0:
            fieldnames[0] = re.sub(r"^#\??\s*", "", fieldnames[0].strip())
        fieldnames = [field.strip().lower() for field in fieldnames]

        validfields = [(i, str(field)) for i, field in enumerate(fieldnames) if field in allfields]
        if not validfields:
            raise RuntimeError("No column has a valid name in the list.")

        for lineno, row in enumerate(reader, start=2):
            if len(row) != len(fieldnames):
                raise RuntimeError(
                    f"line {lineno}: number of fields ({len(row)}) "
                    f"does not agree with what expected ({len(fieldnames)})"
                )
            args: dict[str, Any] = {"lineno": lineno}
            args.update((field, row[i]) for i, field in validfields)
            rects.append(Rect.create(default=default, **args))

    return rects


@contextlib.contextmanager
def open_inputfile(file: Union[str, IO]) -> Generator[IO[bytes], None, None]:
    """
    Open a file with "rb" mode.

    If `file` is a text file object, `file.buffer` will be returned.

    Parameters
    ----------
    file
        A file path or a file object.

    Returns
    -------
    contextmanager
        Context manager.
        When the context is exitted,
        the file is closed if the file has been opened by this function.
        Otherwise, the file is kept open.
    """
    if not isinstance(file, str) and hasattr(file, "read"):
        # This is already a file object
        yield getattr(file, "buffer", file)
    else:
        file = cast(str, file)
        with open(file, "rb") as f:
            yield f


def parse_rerun(s: str) -> str:
    """
    Interpret a string representing a rerun name.

    Parameters
    ----------
    s
        Rerun name.

    Returns
    -------
    rerun
        Rerun name.
    """
    lower = s.lower()
    if lower in available_reruns:
        return lower
    raise ValueError(f"Invalid rerun: {s}")


def parse_type(s: str) -> str:
    """
    Interpret a string representing an image type.

    Parameters
    ----------
    s
        Image type.

    Returns
    -------
    type
        Image type.
    """
    lower = s.lower()
    if lower in available_types:
        return lower
    raise ValueError(f"Invalid type: {s}")


def parse_tract_opt(s: Union[str, int, None]) -> int:
    """
    Interpret a string (etc) representing a tract.

    Parameters
    ----------
    s
        Tract.
        This argument may be `ANYTRACT`.

    Returns
    -------
    tract
        Tract in an integer.
    """
    if s is None:
        return ANYTRACT
    if isinstance(s, int):
        return s

    s = s.lower()
    if s == "any":
        return ANYTRACT
    return int(s)


def parse_bool(s: Union[str, bool]) -> bool:
    """
    Interpret a string (etc) representing a boolean value.

    Parameters
    ----------
    s
        A string (etc) representing a boolean value.

    Returns
    -------
    b
        True or False.
    """
    if isinstance(s, bool):
        return s

    if isinstance(s, np.bool):
        return s

    return {
        "false": False,
        "f": False,
        "no": False,
        "n": False,
        "off": False,
        "0": False,
        "true": True,
        "t": True,
        "yes": True,
        "y": True,
        "on": True,
        "1": True,
    }[s.lower()]


def parse_longitude(s: Union[str, float]) -> float:
    """
    Interpret a longitude.

    Parameters
    ----------
    s
        A string representing a longitude,
        or a float value in degrees.

    Returns
    -------
    longitude
        Degrees.
    """
    type, value = _parse_angle(s)
    if type == "sex":
        return 15 * value
    else:
        return value


def parse_latitude(s: Union[str, float]) -> float:
    """
    Interpret a latitude.

    Parameters
    ----------
    s
        A string representing a latitude,
        or a float value in degrees.

    Returns
    -------
    latitude
        Degrees.
    """
    type, value = _parse_angle(s)
    return value


def parse_degree(s: Union[str, float]) -> float:
    """
    Interpret an angle, which is in degrees by default.

    Parameters
    ----------
    s
        A string representing an angle,
        or a float value in degrees.

    Returns
    -------
    angle
        Degrees.
    """
    type, value = _parse_angle(s)
    return value


def _parse_angle(s: Union[str, float]) -> tuple[str, float]:
    """
    Interpret an angle.

    Parameters
    ----------
    s
        A string representing an angle.

    Returns
    -------
    type
      - "bare"
        `s` did not have its unit.
        What `angle` means must be decided by the caller.
      - "sex"
        `s` was in "99:99:99.999". It may be hours or degrees.
        What `angle` means must be decided by the caller.
      - "deg"
        `angle` is in degrees.

    angle
        a float value
    """
    try:
        if isinstance(s, (float, int)):
            return "bare", float(s)

        s = re.sub(r"\s", "", s).lower()
        m = re.match(
            r"\A(.+)(deg|degrees?|amin|arcmin|arcminutes?|asec|arcsec|arcseconds?|rad|radians?)\Z", s
        )
        if m:
            value, unit = m.groups()
            return "deg", float(value) * _angle_units[unit]

        m = re.match(r"\A([+\-]?)([0-9].*)d([0-9].*)m([0-9].*)s\Z", s)
        if m:
            sign_s, degrees, minutes, seconds = m.groups()
            sign = -1.0 if sign_s == "-" else 1.0
            return "deg", sign * (float(seconds) / 3600 + float(minutes) / 60 + float(degrees))

        m = re.match(r"\A([+\-]?)([0-9].*)h([0-9].*)m([0-9].*)s\Z", s)
        if m:
            sign_s, hours, minutes, seconds = m.groups()
            sign = -1.0 if sign_s == "-" else 1.0
            return "deg", 15.0 * sign * (float(seconds) / 3600 + float(minutes) / 60 + float(hours))

        m = re.match(r"\A([+\-]?)([0-9].*):([0-9].*):([0-9].*)\Z", s)
        if m:
            sign_s, degrees, minutes, seconds = m.groups()
            sign = -1.0 if sign_s == "-" else 1.0
            return "sex", sign * (float(seconds) / 3600 + float(minutes) / 60 + float(degrees))

        return "bare", float(s)

    except Exception:
        raise ValueError(f"Cannot interpret angle: '{s}'") from None


_angle_units = {
    "deg": 1.0,
    "degree": 1.0,
    "degrees": 1.0,
    "amin": 1.0 / 60,
    "arcmin": 1.0 / 60,
    "arcminute": 1.0 / 60,
    "arcminutes": 1.0 / 60,
    "asec": 1.0 / 3600,
    "arcsec": 1.0 / 3600,
    "arcsecond": 1.0 / 3600,
    "arcseconds": 1.0 / 3600,
    "rad": 180 / math.pi,
    "radian": 180 / math.pi,
    "radians": 180 / math.pi,
}


def parse_filter(s: str) -> str:
    """
    Interpret a filter name.

    Parameters
    ----------
    s
        A string representing a filter.
        This may be an alias of a filter name.
        (Like "g" for "HSC-G")

    Returns
    -------
    filter
        A filter name.
    """
    if s in _all_filters:
        return s

    for physicalname, info in _all_filters.items():
        if s in info["alias"]:
            return physicalname

    raise ValueError(f"filter '{s}' not found.")


_all_filters = dict(
    [
        ("HSC-G", {"alias": {"W-S-G+", "g"}, "display": "g"}),
        ("HSC-R", {"alias": {"W-S-R+", "r"}, "display": "r"}),
        ("HSC-I", {"alias": {"W-S-I+", "i"}, "display": "i"}),
        ("HSC-Z", {"alias": {"W-S-Z+", "z"}, "display": "z"}),
        ("HSC-Y", {"alias": {"W-S-ZR", "y"}, "display": "y"}),
        ("IB0945", {"alias": {"I945"}, "display": "I945"}),
        ("NB0387", {"alias": {"N387"}, "display": "N387"}),
        ("NB0400", {"alias": {"N400"}, "display": "N400"}),
        ("NB0468", {"alias": {"N468"}, "display": "N468"}),
        ("NB0515", {"alias": {"N515"}, "display": "N515"}),
        ("NB0527", {"alias": {"N527"}, "display": "N527"}),
        ("NB0656", {"alias": {"N656"}, "display": "N656"}),
        ("NB0718", {"alias": {"N718"}, "display": "N718"}),
        ("NB0816", {"alias": {"N816"}, "display": "N816"}),
        ("NB0921", {"alias": {"N921"}, "display": "N921"}),
        ("NB0926", {"alias": {"N926"}, "display": "N926"}),
        ("NB0973", {"alias": {"N973"}, "display": "N973"}),
        ("NB1010", {"alias": {"N1010"}, "display": "N1010"}),
        ("ENG-R1", {"alias": {"109", "r1"}, "display": "r1"}),
        ("PH", {"alias": {"PH"}, "display": "PH"}),
        ("SH", {"alias": {"SH"}, "display": "SH"}),
        ("MegaCam-u", {"alias": {"u2"}, "display": "MegaCam-u"}),
        ("MegaCam-uS", {"alias": {"u1"}, "display": "MegaCam-uS"}),
        ("VIRCAM-H", {"alias": {"Hvir", "hvir"}, "display": "VIRCAM-H"}),
        ("VIRCAM-J", {"alias": {"Jvir", "jvir"}, "display": "VIRCAM-J"}),
        ("VIRCAM-Ks", {"alias": {"Ksvir", "ksvir"}, "display": "VIRCAM-Ks"}),
        ("VIRCAM-NB118", {"alias": {"NB118vir", "n118vir"}, "display": "VIRCAM-NB118"}),
        ("VIRCAM-Y", {"alias": {"Yvir", "yvir"}, "display": "VIRCAM-Y"}),
        ("WFCAM-H", {"alias": {"Hwf", "hwf"}, "display": "WFCAM-H"}),
        ("WFCAM-J", {"alias": {"Jwf", "jwf"}, "display": "WFCAM-J"}),
        ("WFCAM-K", {"alias": {"Kwf", "kwf"}, "display": "WFCAM-K"}),
    ]
)


def parse_filter_opt(s: Optional[Union[str, list]]) -> Union[str, list]:
    """
    Interpret a filter name.
    The argument may be `ALLFILTERS`.or None
    (both have the same meaning).

    Parameters
    ----------
    s
        A string representing a filter.
        This may be an alias of a filter name.
        (Like "g" for "HSC-G")
        Or it may be `ALLFILTERS`.
        If `s` is None, it has the same meaning as `ALLFILTERS`.

    Returns
    -------
    filter
        A filter name.
        This can be `ALLFILTERS`
    """
    if s is None:
        return ALLFILTERS
    if isinstance(s, list):
        return [parse_filter(filter) for filter in s]
    if s.lower() == ALLFILTERS:
        return ALLFILTERS
    return parse_filter(s)


@export
def download(
    rects: Union[Rect, list[Rect]],
    user: Optional[str] = None,
    password: Optional[str] = None,
    *,
    onmemory: bool = True,
    **kwargs__download,
) -> Union[list, list[list], None]:
    """
    Cut `rects` out of the sky.

    Parameters
    ----------
    rects : Union[Rect, list[Rect]]
        A `Rect` object or a list of `Rect` objects to download
    user : Optional[str]
        Username. If None, it will be asked interactively.
    password : Optional [str]
        Password. If None, it will be asked interactively.
    onmemory, optional : bool
        Return `datalist` on memory.
        If `onmemory` is False, downloaded cut-outs are written to files.
    kwargs__download: dict
        Additional keyword args are passed through to _download and eventually to _download_chunk() and
        urllib.request.urlopen.

        Some important (but entirely optional!) keyword args processed later in the download callstack are
        listed below. Anything urllib.request.urlopen will accept is fair game too!

            chunksize : int
                The number of rects to include in a single http request. Default 990 rects. See _download()
                for greater detail.
            retries : int
                The number of times to retry a failed http request before moving on or attempting to trigger
                failed_chunk_handler. Default is 3. See _download() for greater detail.
            retrywait : int
                Base seconds to wait between retries. Acts as base for exponential retry formula of
                base * (2**attempt_num) seconds. Default is 30 s. See _download() for greater detail.
            failed_chunk_hook : Callable[[list[Rect], Exception, int], Any]
                Hook which is called every time a chunk fails `retries` time. The arguments to the hook are
                the rects in the failed chunk, the exception encountered while making the last request, and
                the number of attempts. See _download() for greater detail.
            request_hook : Callable[[urllib.request.Request, datetime.datetime, datetime.datetime, int, int]
                Optional hook called on every request. See _download_chunk() for greater detail

    Returns
    -------
    datalist : Union[list, list[list], None]
        If onmemory == False, `datalist` is None.
        If onmemory == True:
          - If `rects` is a simple `Rect` object,
            `datalist[j]` is a tuple `(metadata: dict, data: bytes)`.
            This is a list because there may be more than one file
            for a single `Rect` (if, say, filter==ALLFILTERS).
            This list may also be empty, which means no data was found.
          - If `rects` is a list of `Rect` objects,
            `datalist[i]` corresponds to `rects[i]`, and
            `datalist[i][j]` is a tuple `(metadata: dict, data: bytes)`.
    """
    isscalar = isinstance(rects, Rect)
    if isscalar:
        rects = [cast(Rect, rects)]
    rects = cast(list[Rect], rects)

    ret = _download(rects, user, password, onmemory=onmemory, **kwargs__download)
    if isscalar and onmemory:
        ret = cast(list[list], ret)
        return ret[0]

    return ret


def _download(
    rects: list[Rect],
    user: Optional[str],
    password: Optional[str],
    *,
    onmemory: bool,
    chunksize: int = 990,
    retries: int = 3,
    retrywait: int = 30,
    **kwargs__download_chunk,
) -> Optional[list[list]]:
    """
    Cut `rects` out of the sky. Implements configurable request size, retries and exponential backoff.

    When `onmemory == False` this function saves snapshots to the current working directory.
    onmemory has no default and must be specified.

    When `onmemory == False ` and the current working directory contains resume data, and
    `resume = True` (default: False), then the download will be resumed.

    Parameters
    ----------
    rects
        A list of `Rect` objects
    user
        Username. If None, it will be asked interactively.
    password
        Password. If None, it will be asked interactively.
    onmemory: bool
        Return `datalist` on memory.
        If `onmemory` is False, downloaded cut-outs are written to files in the current working directory.
    chunksize: int, optional
        Number of cutout lines to pack into a single request. Defaults to 990 if unspecified.
    retries: int, optional
        Number of attempts to make to fetch each chunk. Defaults to 3 if unspecified.
    retrywait: int, optional
        Base number of seconds to wait between retries. Retry waits are computed using an exponential backoff
        where the retry time for attempts is calculated as retrywait * (2 ** attempt) seconds , with attempt=0
        for the first wait.
    kwargs__download_chunk: dict, optional
        Additional keyword args are passed through to _download_chunk()

    Returns
    -------
    datalist
        If onmemory == False, `datalist` is None.
        If onmemory == True,
        `datalist[i]` corresponds to `rects[i]`, and
        `datalist[i][j]` is a tuple `(metadata: dict, data: bytes)`.
    """
    if not rects:
        return [] if onmemory else None

    for rect in rects:
        if not rect.iscomplete():
            raise RuntimeError(f"'ra', 'dec', 'sw', and 'sh' must be specified: {rect}")

    exploded_rects: list[tuple[Rect, int]] = []
    for index, rect in enumerate(rects):
        exploded_rects.extend((r, index) for r in rect.explode())

    # Sort the rects so that the server can use cache
    # as frequently as possible.
    # We will later use `index` to sort them back.
    exploded_rects.sort()

    if not user:
        user = input("username? ").strip()
        if not user:
            raise RuntimeError("User name is empty.")

    if not password:
        password = getpass.getpass(prompt="password? ")
        if not password:
            raise RuntimeError("Password is empty.")

    datalist: list[tuple[int, dict, bytes]] = []

    # Chunk loop
    for i in range(0, len(exploded_rects), chunksize):
        # Retry loop
        for attempt in range(0, retries):
            try:
                ret = _download_chunk(
                    exploded_rects[i : i + chunksize],
                    user,
                    password,
                    onmemory=onmemory,
                    **kwargs__download_chunk,
                )
                break
            except KeyboardInterrupt:
                logger.critical("Keyboard Interrupt recieved.")
                raise
            except Exception as exception:
                # Humans count attempts from 1, this loop counts from zero.
                logger.warning(
                    f"Attempt {attempt + 1} of {retries} to request rects [{i}:{i + chunksize}] has error:"
                )
                logger.warning(exception)

                # If the final attempt on this chunk fails, we move on.
                if attempt + 1 == retries:
                    break
                # Otherwise wait for exponential backoff and try again
                else:
                    backoff = retrywait * (2**attempt)
                    if backoff != 0:
                        logger.info(f"Retrying in {backoff} seconds... ")
                        time.sleep(backoff)
                        logger.info("Retrying now")
                    continue
        if onmemory:
            datalist += cast(list, ret)

    if onmemory:
        returnedlist: list[list[tuple[dict, bytes]]] = [[] for i in range(len(rects))]
        for index, metadata, data in datalist:
            returnedlist[index].append((metadata, data))

    return returnedlist if onmemory else None


def _download_chunk(
    rects: list[tuple[Rect, Any]],
    user: str,
    password: str,
    manifest: Optional[dict[Rect, str]],
    *,
    onmemory: bool,
    request_hook: Optional[
        Callable[[urllib.request.Request, datetime.datetime, datetime.datetime, int, int], Any]
    ] = None,
    rect_hook: Optional[Callable[[Rect, str], Any]] = None,
    **kwargs_urlopen,
) -> Optional[list]:
    """
    Cut `rects` out of the sky.

    Parameters
    ----------
    rects
        A list of `(Rect, Any)`.
        The length of this list must be smaller than the server's limit.
        Each `Rect` object must be explode()ed beforehand.
        The `Any` value attached to each `Rect` object is a marker.
        The marker is used to indicate the `Rect` in the returned list.
    user
        Username.
    password
        Password.
    manifest
        Dictionary from Rect to filename. If Provided, this function will fill in as it downloads.
        If download of a file fails, the file's entry will read "Attempted" for the filename.
    onmemory
        Return `datalist` on memory.
        If `onmemory` is False, downloaded cut-outs are written to files.
    request_hook
        Function that is called with the response of all requests made
        Intended to support bandwidth instrumentation.
    rect_hook
        Function to be called on every rectangle downloaded. The callback recieves the rect and the filename
        as arguments
    kwargs_urlopen
        Additional keyword args are passed through to urllib.request.urlopen

    Returns
    -------
    datalist
        If onmemory == False, `datalist` is None.
        If onmemory == True,
        each element is a tuple `(marker: Any, metadata: dict, data: bytes)`.
        For `marker`, see the comment for the parameter `rects`.
        Two or more elements in this list may result
        from a single `Rect` object.
    """
    fields = list(_format_rect_member.keys())
    coordlist = [f"#? {' '.join(fields)}"]
    for rect, _ in rects:
        coordlist.append(" ".join(_format_rect_member[field](getattr(rect, field)) for field in fields))

    boundary = "Boundary"
    header = f'--{boundary}\r\nContent-Disposition: form-data; name="list"; filename="coordlist.txt"\r\n\r\n'
    footer = f"\r\n--{boundary}--\r\n"

    data = (header + "\n".join(coordlist) + footer).encode("utf-8")
    secret = base64.standard_b64encode(f"{user}:{password}".encode()).decode("ascii")

    req = urllib.request.Request(
        api_url.rstrip("/") + "/cgi-bin/cutout",
        data=data,
        headers={
            "Authorization": f"Basic {secret}",
            "Content-Type": f'multipart/form-data; boundary="{boundary}"',
        },
        method="POST",
    )

    returnedlist = []

    # Set timeout to 1 hour if no timout was set higher up
    kwargs_urlopen.setdefault("timeout", 3600)

    # Set all manifest entries to indicate an attempt was made.
    if manifest is not None:
        for rect, _ in rects:
            manifest[rect] = "Attempted"

    with get_connection_semaphore():
        request_started = datetime.datetime.now()
        with urllib.request.urlopen(req, **kwargs_urlopen) as fin:
            response_started = datetime.datetime.now()
            response_size = 0
            with tarfile.open(fileobj=fin, mode="r|") as tar:
                for info in tar:
                    fitem = tar.extractfile(info)
                    response_size += info.size
                    if fitem is None:
                        continue
                    with fitem:
                        metadata = _tar_decompose_item_name(info.name)
                        rect, index = rects[metadata["lineno"] - 2]
                        # Overwrite metadata's lineno (= lineno in this chunk)
                        # with rect's lineno (= global lineno)
                        # for fear of confusion.
                        metadata["lineno"] = rect.lineno
                        # Overwrite rect's tract (which may be ANYTRACT)
                        # with metadata's tract (which is always a valid value)
                        # for fear of confusion.
                        rect.tract = metadata["tract"]
                        metadata["rect"] = rect
                        if onmemory:
                            returnedlist.append((index, metadata, fitem.read()))
                        else:
                            filename = make_filename(metadata)
                            dirname = os.path.dirname(filename)
                            if dirname:
                                os.makedirs(dirname, exist_ok=True)
                            with open(filename, "wb") as fout:
                                _splice(fitem, fout)
                            if rect_hook:
                                rect_hook(rect, filename)
                            if manifest is not None:
                                manifest[rect] = filename
            if request_hook:
                request_hook(req, request_started, response_started, response_size, len(rects))

    return returnedlist if onmemory else None


_format_rect_member: dict[str, Callable[[str], Any]] = {
    "rerun": str,
    "type": str,
    "filter": str,
    "tract": lambda x: ("any" if x == ANYTRACT else str(x)),
    "ra": lambda x: f"{x:.16e}deg",
    "dec": lambda x: f"{x:.16e}deg",
    "sw": lambda x: f"{x:.16e}deg",
    "sh": lambda x: f"{x:.16e}deg",
    "image": lambda x: ("true" if x else "false"),
    "mask": lambda x: ("true" if x else "false"),
    "variance": lambda x: ("true" if x else "false"),
}


def _tar_decompose_item_name(name: str) -> dict:
    """
    Get a metadata dictionary by decomposing an item name in a tar file.

    Parameters
    ----------
    name
        The name of an item in a tar file returned by the server.

    Returns
    -------
    metadata
        A dictionary that has the following keys:
          - "lineno": Line number (starting with 2).
          - "type": "coadd", "coadd/bg", or "warp".
          - "filter": Filter name.
          - "tract": Tract number.
          - "rerun": Rerun name.
          - "visit": (warp only) Visit number.
    """
    m = re.fullmatch(
        r"arch-[0-9]+-[0-9]+/(?P<lineno>[0-9]+)-(?P<type>cutout|coadd\+bg)-(?P<filter>[^/]+)-(?P<tract>[0-9]+)-(?P<rerun>[^/]+)\.fits",
        name,
    )
    if m:
        metadata: dict[str, Any] = m.groupdict()
        metadata["lineno"] = int(metadata["lineno"])
        metadata["type"] = {"cutout": "coadd", "coadd+bg": "coadd/bg"}[metadata["type"]]
        metadata["tract"] = int(metadata["tract"])
        return metadata

    m = re.fullmatch(
        r"arch-[0-9]+-[0-9]+/(?P<lineno>[0-9]+)-warps-(?P<filter>[^/]+)-(?P<tract>[0-9]+)-(?P<rerun>[^/]+)/warp-(?P<visit>[0-9]+)\.fits",
        name,
    )
    if m:
        metadata: dict[str, Any] = m.groupdict()  # type: ignore[no-redef]
        metadata["lineno"] = int(metadata["lineno"])
        metadata["type"] = "warp"
        metadata["tract"] = int(metadata["tract"])
        metadata["visit"] = int(metadata["visit"])
        return metadata

    raise ValueError("File name not interpretable")


@export
def make_filename(metadata: dict) -> str:
    """
    Make a filename from `metadata` that is returned by `download(onmemory=True)`.

    Parameters
    ----------
    metadata
        A metadata dictionary.

    Returns
    -------
    filename
        File name.
    """
    rect = metadata["rect"]
    args = dataclasses.asdict(rect)
    type = args["type"]
    if type == "warp":
        args["name"] = f"{args['name']}_{metadata['visit']:06d}"
    if type == "coadd/bg":
        args["type"] = "coadd+bg"

    name = args.pop("name")
    return str(name.format(**args)) + ".fits"


def _splice(fin: IO[bytes], fout: IO[bytes]):
    """
    Read from `fin` and write to `fout` until the end of file.

    Parameters
    ----------
    fin
        Input file.
    fout
        Output file.
    """
    buffer = memoryview(bytearray(10485760))
    while True:
        n = fin.readinto(buffer)  # type: ignore[attr-defined]
        if n <= 0:
            break
        fout.write(buffer[:n])


class Semaphore:
    """
    Named semaphore.

    This semaphore can be shared by multiple machines
    that mount a shared NFS.

    It is guaranteed that the semaphore locked by this process
    is automatically unlocked as soon as this process terminates.

    Parameters
    ----------
    init_num
        Initial semaphore count.
        Zero or negative values mean infinity.
    path
        Path of the semaphore.
        E.g. "/tmp/semaphore-cutout", "/share/data/sem_cutout" etc.
        Be warned that many files will be made under this path (directory).
    """

    def __init__(self, init_num: int, path: str):
        self.init_num = init_num
        self.path = path
        self.fd = -1

        if init_num <= 0:
            return

        os.makedirs(path, exist_ok=True)

    def __del__(self):
        self.unlock()

    def __enter__(self) -> "Semaphore":
        self.lock()
        return self

    def __exit__(self, *args):
        self.unlock()

    def lock(self):
        """
        Wait the semaphore.
        """
        if self.init_num <= 0:
            return
        if self.fd != -1:
            return

        fd = -1

        try:
            while True:
                for i in range(self.init_num):
                    fd = os.open(os.path.join(self.path, f"{i}.sem"), os.O_RDWR | os.O_CREAT, 0o644)
                    try:
                        os.lockf(fd, os.F_TLOCK, 0)
                        self.fd = fd
                        fd = -1
                        return
                    except OSError as e:
                        if e.errno not in (errno.EACCES, errno.EAGAIN):
                            raise

                    os.close(fd)
                    fd = -1

                time.sleep(0.01 * self.init_num)
        finally:
            if fd != -1:
                os.close(fd)

    def unlock(self):
        """
        Signal the semaphore.
        """
        if self.init_num <= 0:
            return
        if self.fd != -1:
            tempfd = self.fd
            self.fd = -1
            os.close(tempfd)


_sem_connections = Semaphore(0, "")


@export
def set_max_connections(num: int, semaphore_path: str = ""):
    """
    Set maximum number of connections.

    Parameters
    ----------
    num
        Number of connections. (Limitless if zero or negative)
    semaphore_path
        Path to the semaphore.
        The default name is `/tmp/$(id -u -n)-downloadCutout`.
    """
    global _sem_connections

    if num > 0 and not semaphore_path:
        semaphore_path = os.path.join(tempfile.gettempdir(), f"{getpass.getuser()}-downloadCutout")

    _sem_connections = Semaphore(num, semaphore_path)


def get_connection_semaphore() -> Semaphore:
    """
    Get the semaphore configured by `set_max_connections()`

    Returns
    -------
    semaphore
        Semaphore configured by `set_max_connections()`
    """
    return _sem_connections


if __name__ == "__main__":
    main()



"""
Hyrax has several built-in datasets that you can use for astronomical data. For many uses, these datasets
can be configured out-of-the box for a given project.

:doc:`FitsImageDataSet <fits_image_dataset/index>` is a generic container for fits image cutout data
indexed by a user-provided catalog file. It attempts to cover common usage paradigms such as multiple images
of the same object differentiated by telescope filter; however, extending the class as a custom dataset
may be more well fit to advanced usage.

:doc:`LSSTDataset <lsst_dataset/index>` Is a alpha-quality container for LSST cutout images, currently
limited to ``deep_coadd`` type images, and restricted to run only on a Rubin observatory RSP environment
where `LSST Pipeline <https://pipelines.lsst.io/>`_ tools and a
`data butler <https://pipelines.lsst.io/modules/lsst.daf.butler/index.html>`_ with the appropriate images
are available.

:doc:`DownloadedLSSTDataset <downloaded_lsst_dataset/index>` is a subclass of LSSTDataset that generates
cutouts from the butler and saves them as ``.pt`` files on first access. On subsequent access,
it loads the cutouts directly from these files, which can significantly speed up data loading times.
It inherits from LSSTDataset to access the data butler and catalog functionality.

:doc:`HSCDataSet <hsc_data_set/index>` Works similarly to FitsImageDataSet, but is specialized to
`Hyper Suprime-Cam (HSC) <https://hsc-release.mtk.nao.ac.jp/doc/index.php/data/>`_ cutout images downloaded
with the hyrax ``download`` verb. It contains additional integrity checks and is tightly integrated with
the ``download`` and ``rebuild_manifest`` verbs. In future this class and the downloader may become a
separate package.

:doc:`HyraxCifarDataSet <hyrax_cifar_data_set/index>` and
:doc:`HyraxCifarIterableDataSet <hyrax_cifar_data_set/index>` give access to the standard
`CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ labeled image dataset, automatically downloading the
dataset if it is not present. These datasets are useful for testing hyrax and occasionally individual models,
but they are not astronomical datasets.

:doc:`HyraxRandomDataset <random/hyrax_random_dataset/index>` and
:doc:`HyraxRandomIterableDataset <random/hyrax_random_dataset/index>` are utility datasets that
generate random data with a specific shape.
These datasets make it easy to test new models with simple random data.
They are highly configurable such that it's possible to simulate input data for models that
are under development.

Each of these datasets can be used a starting point for a Custom Dataset by inheriting your custom dataset
from e.g. `FitsImageDataSet`, or you can make an entirely custom dataset following the
:ref:`custom dataset instructions <custom-dataset-instructions>` and/or
:doc:`custom dataset example notebook </pre_executed/custom_dataset>`.

The remaining classes in this module exist primarily for Hyrax interface purposes:

:doc:`InferenceDataset <inference_dataset/index>` is a dataset class that represents an ``infer`` or ``umap``
result, and may be returned from those verbs to provide data access

:doc:`HyraxDataset <data_set_registry/index>` is a base class for all datasets in Hyrax and must be within
the inheretence hierarchy of all custom datasets. It is not usable on it's own, but provides various fall-back
functionality to make custom datasets easier to write. See the
:ref:`custom dataset instructions <custom-dataset-instructions>` and
:doc:`example notebook </pre_executed/custom_dataset>` for more information.

"""

# Remove import sorting, these are imported in the order written so that
# autoapi docs are generated with ordering controlled below.
# ruff: noqa: I001
from .fits_image_dataset import FitsImageDataSet
from .lsst_dataset import LSSTDataset
from .downloaded_lsst_dataset import DownloadedLSSTDataset
from .hsc_data_set import HSCDataSet
from .hyrax_cifar_data_set import HyraxCifarDataSet, HyraxCifarIterableDataSet
from .random.hyrax_random_dataset import (
    HyraxRandomDataset,
    HyraxRandomIterableDataset,
    HyraxRandomDatasetBase,
)
from .inference_dataset import InferenceDataSet
from .data_set_registry import HyraxDataset
from .hyrax_cifar_data_set import HyraxCifarBase

__all__ = [
    "HyraxCifarDataSet",
    "FitsImageDataSet",
    "HyraxCifarIterableDataSet",
    "HSCDataSet",
    "InferenceDataSet",
    "HyraxDataset",
    "LSSTDataset",
    "DownloadedLSSTDataset",
    "HyraxCifarBase",
    "HyraxRandomDataset",
    "HyraxRandomIterableDataset",
    "HyraxRandomDatasetBase",
]



import numpy as np
import pytest
from torch import any, from_numpy, isnan, tensor

import hyrax
from hyrax.data_sets.random.hyrax_random_dataset import HyraxRandomDataset
from hyrax.pytorch_ignite import _handle_nans


class RandomNaNDataset(HyraxRandomDataset):
    """Dataset yielding pairs of random numbers. Requires a seed to emulate
    static data on the filesystem between instantiations"""

    def __init__(self, config):
        super().__init__(config)

    def __getitem__(self, idx):
        return from_numpy(self.data[idx])


@pytest.fixture(scope="function", params=["RandomNaNDataset", "HyraxRandomDataset"])
def loopback_hyrax_nan(tmp_path_factory, request):
    """This generates a loopback hyrax instance
    which is configured to use the loopback model
    and a simple dataset yielding random numbers
    """
    results_dir = tmp_path_factory.mktemp("loopback_hyrax_nan")

    h = hyrax.Hyrax()
    h.config["model"]["name"] = "HyraxLoopback"
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 5
    h.config["general"]["results_dir"] = str(results_dir)

    h.config["general"]["dev_mode"] = True
    h.config["data_set"]["name"] = request.param
    h.config["data_set.random_dataset"]["size"] = 20
    h.config["data_set.random_dataset"]["seed"] = 0
    h.config["data_set.random_dataset"]["shape"] = [2, 3]
    h.config["data_set.random_dataset"]["number_invalid_values"] = 40
    h.config["data_set.random_dataset"]["invalid_value_type"] = "nan"

    h.config["data_set"]["validate_size"] = 0.2
    h.config["data_set"]["test_size"] = 0.2
    h.config["data_set"]["train_size"] = 0.6

    weights_file = results_dir / "fakeweights"
    with open(weights_file, "a"):
        pass
    h.config["infer"]["model_weights_file"] = str(weights_file)

    dataset = h.prepare()
    return h, dataset


def test_nan_handling(loopback_hyrax_nan):
    """
    Test that default nan handling removes nans
    """
    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = "quantile"

    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(item["image"])) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    for result in inference_results:
        assert not any(isnan(result))


def test_nan_handling_zero_values(loopback_hyrax_nan):
    """
    Test that zero nan handling removes nans
    """
    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = "zero"

    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(item["image"])) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    for result in inference_results:
        assert not any(isnan(result))


def test_nan_handling_off(loopback_hyrax_nan):
    """
    Test that when nan handling is off nans appear in output
    """
    h, dataset = loopback_hyrax_nan

    h.config["data_set"]["nan_mode"] = False
    inference_results = h.infer()

    if isinstance(dataset[0], dict):
        original_nans = tensor([any(isnan(item["image"])) for item in dataset])
    else:
        original_nans = tensor([any(isnan(item)) for item in dataset])
    assert any(original_nans)

    result_nans = tensor([any(isnan(item)) for item in inference_results])
    assert any(result_nans)


def test_nan_handling_off_returns_input(loopback_hyrax_nan):
    """Ensure that when nan_mode is False, that the original values passed to
    _handle_nans are returned unchanged."""

    def to_tensor(data_dict):
        if "image" in data_dict and "label" in data_dict:
            image = data_dict["image"]
            label = data_dict["label"]
            return (image, label)

    h, dataset = loopback_hyrax_nan
    h.config["data_set"]["nan_mode"] = False

    sample_data = dataset[0]

    # If the sample data is a dictionary, convert it to a tuple
    if isinstance(sample_data, dict):
        sample_data = to_tensor(sample_data)

    output = _handle_nans(sample_data, h.config)

    # If the sample was a tuple, check all the elements
    if isinstance(sample_data, tuple):
        assert np.all(np.isclose(output[0], sample_data[0], equal_nan=True))
        assert output[1] == sample_data[1]
    else:
        assert np.all(np.isclose(output, sample_data, equal_nan=True))



import argparse
import sys
from importlib.metadata import version

from hyrax import Hyrax
from hyrax.verbs import all_verbs, fetch_verb_class, is_verb_class


def main():
    """Primary entry point for the Hyrax CLI. This handles dispatching to the various
    Hyrax actions and returning a result.
    """

    description = "Hyrax CLI"
    epilog = "Hyrax is the Framework for Image-Based Anomaly Detection"

    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    _add_major_arguments(parser)

    # cut off "usage: " from beginning and "\n" from end so we get an invocation
    # which subcommand parsers can add to appropriately.
    subparser_usage_prefix = parser.format_usage()[7:-1]
    subparsers = parser.add_subparsers(title="Verbs:", required=False)

    # Add a subparser for every verb, (whether defined by function or class)
    for cli_name in all_verbs():
        subparser_kwargs = {}

        if is_verb_class(cli_name):
            verb_class = fetch_verb_class(cli_name)
            subparser_kwargs = verb_class.add_parser_kwargs

        verb_parser = subparsers.add_parser(
            cli_name, prog=subparser_usage_prefix + " " + cli_name, **subparser_kwargs
        )

        if is_verb_class(cli_name):
            verb_class.setup_parser(verb_parser)

        verb_parser.set_defaults(verb=cli_name)
        _add_major_arguments(verb_parser)

    args = parser.parse_args()

    if args.version:
        print(version("hyrax"))
        return

    if not args.verb:
        parser.print_help()
        sys.exit(1)

    hyrax_instance = Hyrax(config_file=args.runtime_config)
    retval = 0
    if is_verb_class(args.verb):
        verb = fetch_verb_class(args.verb)(hyrax_instance.config)
        retval = verb.run_cli(args)
    else:
        getattr(hyrax_instance, args.verb)()

    exit(retval)


def _add_major_arguments(parser):
    parser.add_argument("--version", dest="version", action="store_true", help="Show version")
    parser.add_argument("-c", "--runtime-config", type=str, help="Full path to runtime config file")


if __name__ == "__main__":
    main()



import logging
from argparse import ArgumentParser, Namespace
from typing import Optional

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class Search(Verb):
    """Stub of similarity search"""

    cli_name = "search"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser(parser: ArgumentParser):
        """Stub of parser setup"""
        parser.add_argument("-i", "--image-file", type=str, help="Path to image file", required=True)

    # If both of these move to the verb superclass then a new verb is basically
    #
    # If you want no args, just make the class, define run(self)
    # If you want args
    #     1) write setup_parser (which sets up for ArgumentParser and name/type info for cli run)
    #     2) write run(self, <your args>) to do what you want
    #

    # Should there be a version of this on the base class which uses a dict on the Verb
    # superclass to build the call to run based on what the subclass verb defined in setup_parser
    def run_cli(self, args: Optional[Namespace] = None):
        """Stub CLI implementation"""
        logger.info("Search run from cli")
        if args is None:
            raise RuntimeError("Run CLI called with no arguments.")
        # This is where we map from CLI parsed args to a
        # self.run (args) call.
        return self.run(image_file=args.image_file)

    def run(self, image_file: str):
        """Search for... todo

        Parameters
        ----------
        image_file : str
            _description_
        """
        logger.info(f"Got Image {image_file}")



import logging
from abc import ABC
from typing import Optional

logger = logging.getLogger(__name__)


class Verb(ABC):  # noqa: B024
    """Base class for all hyrax verbs"""

    # Verbs get to define how their parser gets added to the main parser
    # This is given in case verbs do not define any keyword args for
    # subparser.add_parser()
    add_parser_kwargs: dict[str, str] = {}

    def __init__(self, config):
        """
        .. py:method:: __init__

        Overall initialization for all verbs that saves the config
        """
        self.config = config


# Verbs with no class are assumed to have a function in hyrax.py which
# performs their function. All other verbs should be defined by named classes
# in hyrax.verbs and use the @hyrax_verb decorator
VERB_REGISTRY: dict[str, Optional[type[Verb]]] = {
    "train": None,
    "infer": None,
    "download": None,
    "prepare": None,
    "rebuild_manifest": None,
}


def hyrax_verb(cls: type[Verb]) -> type[Verb]:
    """Decorator to register a hyrax verb"""
    from hyrax.plugin_utils import update_registry

    update_registry(VERB_REGISTRY, cls.cli_name, cls)  # type: ignore[attr-defined]
    return cls


def all_verbs() -> list[str]:
    """Returns all verbs that are currently registered"""
    return [verb for verb in VERB_REGISTRY]


def all_class_verbs() -> list[str]:
    """Returns all verbs that are currently registered with a class-based implementation"""
    return [verb for verb in VERB_REGISTRY if VERB_REGISTRY.get(verb) is not None]


def is_verb_class(cli_name: str) -> bool:
    """Returns true if the verb has a class based implementation

    Parameters
    ----------
    cli_name : str
        The name of the verb on the command line interface

    Returns
    -------
    bool
        True if the verb has a class-based implementation
    """
    return cli_name in VERB_REGISTRY and VERB_REGISTRY.get(cli_name) is not None


def fetch_verb_class(cli_name: str) -> Optional[type[Verb]]:
    """Gives the class object for the named verb

    Parameters
    ----------
    cli_name : str
        The name of the verb on the command line interface


    Returns
    -------
    Optional[type[Verb]]
        The verb class or None if no such verb class exists.
    """
    return VERB_REGISTRY.get(cli_name)



# Remove import sorting, these are imported in the order written so that
# autoapi docs are generated with ordering controlled below.
# ruff: noqa: I001
from hyrax.verbs.database_connection import DatabaseConnection
from hyrax.verbs.umap import Umap
from hyrax.verbs.infer import Infer
from hyrax.verbs.visualize import Visualize
from hyrax.verbs.lookup import Lookup
from hyrax.verbs.save_to_database import SaveToDatabase
from hyrax.verbs.verb_registry import Verb
from hyrax.verbs.verb_registry import all_class_verbs, all_verbs, fetch_verb_class, is_verb_class

__all__ = [
    "VERB_REGISTRY",
    "is_verb_class",
    "fetch_verb_class",
    "all_class_verbs",
    "all_verbs",
    "Lookup",
    "Umap",
    "Visualize",
    "Infer",
    "SaveToDatabase",
    "Verb",
    "DatabaseConnection",
]



import logging

from hyrax.pytorch_ignite import setup_dataset

logger = logging.getLogger(__name__)


def run(config):
    """Prepare the dataset for a given model and data loader.

    Parameters
    ----------
    config : dict
        The parsed config file as a nested
        dict
    """

    data_set = setup_dataset(config)

    logger.info("Finished Prepare")
    return data_set



import logging
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional, Union

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class DatabaseConnection(Verb):
    """Verb to insert inference results into a vector database index for fast
    similarity search."""

    cli_name = "database_connection"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser(parser: ArgumentParser):
        """Stub of parser setup"""

        parser.add_argument(
            "-d",
            "--database-dir",
            type=str,
            required=False,
            help="Directory of existing vector database.",
        )

    def run_cli(self, args: Optional[Namespace] = None):
        """Stub CLI implementation"""
        logger.error("Database connection is not supported from the command line.")

    def run(self, database_dir: Optional[Union[Path, str]] = None):
        """Create a connection to the vector database for interactive queries.

        Parameters
        ----------
        database_dir : str or Path, Optional
            The directory containing the database that will be connected to.
            If None, attempt to connect to the most recently created `...-vector-db-...`
            directory. If specified, it can point to either an empty directory
            or a directory containing an existing vector database. If the latter, the
            database will be updated with the new vectors.
        """
        from hyrax.config_utils import find_most_recent_results_dir
        from hyrax.vector_dbs.vector_db_factory import vector_db_factory

        config = self.config

        # Attempt to find the directory containing the vector database. Check for
        # the database_dir argument first, then check the config file for
        # vector_db.vector_db_dir, and finally check for the most recently
        # created vector-db directory.
        vector_db_dir = None
        if database_dir is not None:
            vector_db_dir = database_dir
        elif config["vector_db"]["vector_db_dir"]:
            vector_db_dir = config["vector_db"]["vector_db_dir"]
        else:
            vector_db_dir = find_most_recent_results_dir(config, "vector-db")

        vector_db_path = Path(vector_db_dir).resolve()
        if not vector_db_path.is_dir():
            raise RuntimeError(
                f"Database directory {str(vector_db_path)} does not exist. \
                    Have you run `hyrax.save_to_database(output_dir={vector_db_path})`?"
            )

        # Get the flavor of database (i.e. Chroma, Qdrant, etc) from the config
        # file saved in `vector_db_path`. This ensures that we will use the correct
        # database class when creating the connection.
        db_type = self._get_database_type_from_config(vector_db_path)
        config["vector_db"]["name"] = db_type

        # Create an instance of the vector database class for the connection
        self.vector_db = vector_db_factory(config, context={"results_dir": vector_db_path})
        if self.vector_db is None:
            raise RuntimeError(f"Unable to conenct to the {db_type} database in directory {vector_db_path}")

        return self.vector_db

    def _get_database_type_from_config(self, database_dir: Path):
        """Internal function that will read a config file from a directory and
        return the name of the vector database from it. i.e. "chromadb", "qdrant".

        Parameters
        ----------
        database_dir : Path
            The directory containing the vector database and the config file that
            be used as reference.

        Returns
        -------
        str
            The config value for ["vector_db"]["name"] in the reference config.
        """
        from hyrax.config_utils import ConfigManager

        config_file = database_dir / "runtime_config.toml"
        reference_config = ConfigManager.read_runtime_config(config_filepath=config_file)
        return reference_config["vector_db"]["name"]



# ruff: noqa: D101, D102

# This example model is taken from the PyTorch CIFAR10 tutorial:
# https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html#define-a-convolutional-neural-network
import logging

import torch
import torch.nn as nn
import torch.nn.functional as F  # noqa N812

from .model_registry import hyrax_model

logger = logging.getLogger(__name__)


@hyrax_model
class HyraxCNN(nn.Module):
    """
    This CNN is designed to work with datasets that are prepared with Hyrax's HSC Data Set class.
    """

    def __init__(self, config, shape):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

        self.config = config

    def forward(self, x):
        # This check is inefficient - we assume that the example CNN will be primarily
        # used with the CIFAR10 dataset. During training, the `train_step` method
        # will unpack the tuple and only pass the first element to the `forward` method.
        # But for inference the entire tuple is passed in, so we need to handle
        # both cases.
        if isinstance(x, tuple):
            x, _ = x

        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def train_step(self, batch):
        """This function contains the logic for a single training step. i.e. the
        contents of the inner loop of a ML training process.

        Parameters
        ----------
        batch : tuple
            A tuple containing the inputs and labels for the current batch.

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """
        inputs, labels = batch

        self.optimizer.zero_grad()
        outputs = self(inputs)
        loss = self.criterion(outputs, labels)
        loss.backward()
        self.optimizer.step()
        return {"loss": loss.item()}



import pytest

from hyrax import plugin_utils
from hyrax.models import hyrax_model
from hyrax.models.model_registry import fetch_model_class


def test_import_module_from_string():
    """Test the import_module_from_string function."""
    module_path = "builtins.BaseException"

    model_cls = plugin_utils.import_module_from_string(module_path)

    assert model_cls.__name__ == "BaseException"


def test_import_module_from_string_no_base_module():
    """Test that the import_module_from_string function raises an error when
    the base module is not found."""

    module_path = "nonexistent.BaseException"

    with pytest.raises(ModuleNotFoundError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Module nonexistent not found" in str(excinfo.value)


def test_import_module_from_string_no_submodule():
    """Test that the import_module_from_string function raises an error when
    a submodule is not found."""

    module_path = "builtins.nonexistent.BaseException"

    with pytest.raises(ModuleNotFoundError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Module builtins.nonexistent not found" in str(excinfo.value)


def test_import_module_from_string_no_class():
    """Test that the import_module_from_string function raises an error when
    a class is not found."""

    module_path = "builtins.Nonexistent"

    with pytest.raises(AttributeError) as excinfo:
        plugin_utils.import_module_from_string(module_path)

    assert "Model class Nonexistent not found" in str(excinfo.value)


def test_fetch_model_class():
    """Test the fetch_model_class function."""
    config = {"model": {"name": "builtins.BaseException"}}

    model_cls = fetch_model_class(config)

    assert model_cls.__name__ == "BaseException"


def test_fetch_model_class_no_model():
    """Test that the fetch_model_class function raises an error when no model
    is specified in the configuration."""

    config = {"model": {}}

    with pytest.raises(ValueError) as excinfo:
        fetch_model_class(config)

    assert "Error fetching model class" in str(excinfo.value)


def test_fetch_model_class_no_model_cls():
    """Test that an exception is raised when a non-existent model class is requested."""

    config = {"model": {"name": "builtins.Nonexistent"}}

    with pytest.raises(AttributeError) as excinfo:
        fetch_model_class(config)

    assert "Model class Nonexistent not found" in str(excinfo.value)


def test_fetch_model_class_not_in_registry():
    """Test that an exception is raised when a model is requested that is not in the registry."""

    config = {"model": {"name": "Nonexistent"}}

    with pytest.raises(ValueError) as excinfo:
        fetch_model_class(config)

    assert "Error fetching model class" in str(excinfo.value)


def test_fetch_model_class_in_registry():
    """Test that a model class is returned when it is in the registry."""

    # make a no-op model that will be added to the model registry
    @hyrax_model
    class NewClass:
        pass

    config = {"model": {"name": "NewClass"}}
    model_cls = fetch_model_class(config)

    assert model_cls.__name__ == "NewClass"



# ruff: noqa: D101, D102

import datetime
import logging
import multiprocessing
import os
import re
import resource
from pathlib import Path
from typing import Optional

import numpy as np
from schwimmbad import MultiPool

from hyrax.config_utils import ConfigDict

from .fits_image_dataset import FitsImageDataSet, files_dict

logger = logging.getLogger(__name__)
dim_dict = dict[str, list[tuple[int, int]]]


class HSCDataSet(FitsImageDataSet):
    """Dataset for sets of HSC cutouts created by the ``fibad download`` command."""

    _called_from_test = False

    def __init__(self, config: ConfigDict):
        """
        .. py:method:: __init__

        """
        from hyrax.download import Downloader

        # Note "rebuild_manifest" is not a config, its a hack for rebuild_manifest mode
        # to ensure we don't use the manifest we believe is corrupt.
        rebuild_manifest = config["rebuild_manifest"] if "rebuild_manifest" in config else False  # noqa: SIM401

        # Set the filter catalog
        # If we are in rebuild manifest mode don't use any filter catalog
        if rebuild_manifest:
            config["data_set"]["filter_catalog"] = False
        # If there's no filter catalog, try to use the manifest file if it exists
        elif not config["data_set"]["filter_catalog"]:
            catalog = Path(config["general"]["data_dir"]) / Downloader.MANIFEST_FILE_NAME
            if catalog.exists():
                config["data_set"]["filter_catalog"] = str(catalog.expanduser().resolve())

        self.filters_config = config["data_set"]["filters"] if config["data_set"]["filters"] else None

        super().__init__(config)

    def _read_filter_catalog(self, filter_catalog_path: Optional[Path]):
        from astropy.table import Table

        try:
            retval = super()._read_filter_catalog(filter_catalog_path)
        except RuntimeError:
            # _read_filter_catalog is persnickity about filter_catalog_path.
            # Ignore all of the error checking in there and _parse_filter_catalog
            # will try to recover if the table is malformed/missing.
            retval = None

        if isinstance(retval, Table):
            colnames = retval.colnames
            if (self.filter_column_name not in colnames) ^ (self.filename_column_name not in colnames):
                msg = f"Filter catalog file {filter_catalog_path} provides one of filters or filenames "
                msg += "without the other. Filesystem scan will still occur without both defined."
                logger.warning(msg)

        return retval

    # The main job of this function is to transmute the filter catalog table into
    # the dictionaries that the rest of the class uses.
    #
    # In the HSC case this will also have to do fallback and call
    # _scan_file_dimensions() and/or _scan_file_names() and pass back only the files dict.
    def _parse_filter_catalog(self, table) -> None:
        object_id_missing = self.object_id_column_name not in table.colnames if table is not None else True
        filter_missing = self.filter_column_name not in table.colnames if table is not None else True
        filename_missing = self.filename_column_name not in table.colnames if table is not None else True

        file_scan = table is None or object_id_missing or filter_missing or filename_missing

        object_ids_for_filescan = None
        if not object_id_missing:
            if filter_missing and filename_missing:
                object_ids_for_filescan = list(table[self.object_id_column_name])
            elif filter_missing or filename_missing:
                object_ids_for_filescan = list(set(table[self.object_id_column_name]))

        # Detect the list of filters, but allow config based override
        if file_scan:
            self.files = self._scan_file_names(self.filters_config, object_ids_for_filescan)
            self.dims = self._scan_file_dimensions()

        # Otherwise we have a well formed table
        else:
            # Have the superclass assemble self.files
            self.files = super()._parse_filter_catalog(table)

            # Assemble dims for ourself if the column is available or fallback to self._scan_file_dimensions()
            if "dim" not in table.colnames:
                self.dims = self._scan_file_dimensions()
            else:
                dim_catalog: dim_dict = {}

                for row in table:
                    object_id = str(row[self.object_id_column_name])
                    # filter = row[self.filter_column_name]
                    filename = row[self.filename_column_name]
                    dim = tuple(row["dim"])

                    # Skip over any files that are marked as didn't download or have <1x1 size, removing the
                    # relevant object from the files dict if it exists
                    if filename == "Attempted" or min(dim) < 1:
                        if object_id in self.files:
                            del self.files[object_id]
                        continue

                    # Dimension is optional, insert into dimension catalog.
                    if object_id not in dim_catalog:
                        dim_catalog[object_id] = []
                    dim_catalog[object_id].append(dim)

                self.dims = dim_catalog

        return self.files

    def _set_crop_transform(self):
        from torchvision.transforms.v2 import CenterCrop

        cutout_shape = self.config["data_set"]["crop_to"] if self.config["data_set"]["crop_to"] else None
        self.cutout_shape = self._check_file_dimensions() if cutout_shape is None else cutout_shape
        return CenterCrop(size=self.cutout_shape)

    def _before_preload(self):
        self.filters_ref = (
            list(list(self.files.values())[0]) if self.filters_config is None else self.filters_config
        )
        self.pruned_objects: dict[str, str] = {}
        self._prune_objects(self.filters_ref, self.cutout_shape)

    def _scan_file_names(
        self, filters: Optional[list[str]], filter_obj_ids: Optional[list[str]] = None
    ) -> files_dict:
        """Class initialization helper

        Parameters
        ----------
        filters: list[str], Optional:
            List of filters that we should look for in the data corpus

        filter_obj_ids: list[str], Optional:
            Filter the file scan to only file names which have the provided object IDs, skipping other files
            When not provided, all file names in the configured data directory that match the pattern from
            hyrax download are parsed.

        Returns
        -------
        dict[str,dict[str,str]]
            Nested dictionary where the first level maps object_id -> dict, and the second level maps
            filter_name -> file name. Corresponds to self.files
        """

        logger.info(f"Scanning files in directory {self.path}")

        object_id_regex = r"[0-9]{17}"
        filter_regex = r"HSC-[GRIZY]" if filters is None else "|".join(filters)
        full_regex = f"({object_id_regex})_.*_({filter_regex}).fits"

        files: files_dict = {}
        # Go scan the path for object ID's so we have a list.
        for index, filepath in enumerate(Path(self.path).iterdir()):
            filename = filepath.name

            # If we are filtering based off a user-provided catalog of object ids, Filter out any
            # objects_ids not in the catalog. Do this before regex match for speed of discarding
            # irrelevant files.
            if isinstance(filter_obj_ids, list) and filename[:17] not in filter_obj_ids:
                continue

            m = re.match(full_regex, filename)
            # Skip files that don't allow us to extract both object_id and filter
            if m is None:
                continue

            object_id = m[1]
            filter = m[2]

            if files.get(object_id) is None:
                files[object_id] = {}

            if files[object_id].get(filter) is None:
                files[object_id][filter] = filename
            else:
                msg = f"Duplicate object ID {object_id} detected.\n"
                msg += f"File {filename} conflicts with already scanned file {files[object_id][filter]} "
                msg += "and will not be included in the data set."
                logger.error(msg)

            if index != 0 and index % 1_000_000 == 0:
                logger.info(f"Processed {index} files.")
        else:
            logger.info(f"Processed {index + 1} files")

        return files

    @staticmethod
    def _determine_numprocs() -> int:
        # Figure out how many CPUs we are allowed to use
        cpu_count = None
        sched_getaffinity = getattr(os, "sched_getaffinity", None)

        if sched_getaffinity:
            cpu_count = len(sched_getaffinity(0))
        elif multiprocessing:
            cpu_count = multiprocessing.cpu_count()
        else:
            cpu_count = 1

        # Ideally we would use ~75 processes per CPU to attempt to saturate
        # I/O bandwidth using a small number of CPUs.
        numproc = 1 if HSCDataSet._called_from_test else 75 * cpu_count
        numproc = HSCDataSet._fixup_limit(
            numproc,
            resource.RLIMIT_NOFILE,
            lambda proc: int(4 * proc + 10),
            lambda nofile: int((nofile - 10) / 4),
        )

        numproc = HSCDataSet._fixup_limit(
            numproc, resource.RLIMIT_NPROC, lambda proc: proc, lambda proc: proc
        )
        return numproc

    @staticmethod
    def _fixup_limit(nproc: int, res, est_limit, est_procs) -> int:
        # If launching this many processes would trigger other resource limits, work around them
        limit_soft, limit_hard = resource.getrlimit(res)

        # If we would violate the hard limit, calculate the number of processes that wouldn't
        # violate the limit
        if limit_hard < est_limit(nproc):
            nproc = est_procs(limit_hard)

        # If we would violate the soft limit, attempt to change it, leaving the hard limit alone
        try:
            if limit_soft < est_limit(nproc):
                resource.setrlimit(res, (est_limit(nproc), limit_hard))
        finally:
            # If the change doesn't take, then reduce the number of processes again
            limit_soft, limit_hard = resource.getrlimit(res)
            if limit_soft < est_limit(nproc):
                nproc = est_procs(limit_soft)

        return nproc

    def _scan_file_dimensions(self) -> dim_dict:
        # Scan the filesystem to get the widths and heights of all images into a dict
        logger.info("Scanning for dimensions...")

        # So we can use super() with no args inside the generator expression below
        super_obj = super()
        retval = {}
        with MultiPool(processes=HSCDataSet._determine_numprocs()) as pool:
            args = (
                (object_id, list(super_obj._object_files(object_id)))
                for object_id in self.ids(log_every=1_000_000)
            )
            retval = dict(pool.imap(self._scan_file_dimension, args, chunksize=1000))
        return retval

    @staticmethod
    def _scan_file_dimension(processing_unit: tuple[str, list[str]]) -> tuple[str, list[tuple[int, int]]]:
        object_id, filenames = processing_unit
        return (object_id, [HSCDataSet._fits_file_dims(filepath) for filepath in filenames])

    @staticmethod
    def _fits_file_dims(filepath) -> tuple[int, int]:
        from astropy.io import fits

        try:
            with fits.open(filepath) as hdul:
                return (hdul[1].shape[0], hdul[1].shape[1])
        except OSError:
            return (0, 0)

    def _prune_objects(self, filters_ref: list[str], cutout_shape: Optional[tuple[int, int]]):
        """Class initialization helper. Prunes objects from the list of objects.

        1) Removes any objects which do not have all the filters specified in filters_ref
        2) If a cutout_shape was provided in the constructor, prunes files that are too small
           for the chosen cutout size

        This function deletes from self.files and self.dims via _prune_object

        Parameters
        ----------
        files : dict[str,dict[str,str]]
            Nested dictionary where the first level maps object_id -> dict, and the second level maps
            filter_name -> file name. This is created by _scan_files()

        filters_ref : list[str]
            List of the filter names

        cutout_shape: tuple[int, int]
            Cutout shape tuple provided from constructor
        """
        filters_ref = sorted(filters_ref)
        self.prune_count = 0
        for index, (object_id, filters_unsorted) in enumerate(self.files.items()):
            # Drop objects with missing filters
            filter_intersect = sorted([filter for filter in filters_unsorted if filter in filters_ref])
            if filter_intersect != filters_ref:
                msg = f"HSCDataSet in {self.path} has the wrong group of filters for object {object_id}."
                self._mark_for_prune(object_id, msg)
                logger.info(f"Filters for object {object_id} were {filters_unsorted}")
                logger.debug(f"Reference filters were {filters_ref}")

            elif cutout_shape is not None:
                # Drop objects that can't meet the cutout size provided
                for shape in self.dims[object_id]:
                    if shape[0] < cutout_shape[0] or shape[1] < cutout_shape[1]:
                        msg = f"A file for object {object_id} has shape ({shape[0]}px, {shape[1]}px)"
                        msg += " this is too small for the given cutout size of "
                        msg += f"({cutout_shape[0]}px, {cutout_shape[1]}px)"
                        self._mark_for_prune(object_id, msg)
                        break

                # Drop objects where the cutouts are not the same size
                first_shape = None
                for shape in self.dims[object_id]:
                    first_shape = shape if first_shape is None else first_shape
                    if shape != first_shape:
                        msg = f"The first filter for object {object_id} has a shape of "
                        msg += f"({first_shape[0]}px,{first_shape[1]}px) another filter has shape of"
                        msg += f"({shape[0]}px,{shape[1]}px)"
                        self._mark_for_prune(object_id, msg)
                        break

                # Drop objects where parsing the filenames does not reveal the object IDs
                for filter, filepath in filters_unsorted.items():
                    filename = Path(filepath).name
                    # Check beginning of filename vs object_id
                    if filename[:17] != object_id:
                        msg = f"Filter {filter} for object id {object_id} has filename {filepath} listed"
                        msg += "The filename does not match the object_id, and the filter_catalog or "
                        msg += "manifest is likely corrupt."
                        self._mark_for_prune(object_id, msg)
                        break

            if index != 0 and index % 1_000_000 == 0:
                logger.info(f"Processed {index} objects for pruning")
        else:
            logger.info(f"Processed {index + 1} objects for pruning")

        # Prune marked objects
        for object_id, reason in self.pruned_objects.items():
            self._prune_object(object_id, reason)

        # Log about the pruning process
        pre_prune_object_count = len(self.files) + self.prune_count
        prune_fraction = self.prune_count / pre_prune_object_count
        if prune_fraction > 0.05:
            logger.error("Greater than 5% of objects in the data directory were pruned.")
        elif prune_fraction > 0.01:
            logger.warning("Greater than 1% of objects in the data directory were pruned.")

        if self.prune_count > 0:
            logger.info(f"Pruned {self.prune_count} out of {pre_prune_object_count} objects")

    def _mark_for_prune(self, object_id, reason):
        self.pruned_objects[object_id] = reason

    def _prune_object(self, object_id, reason: str):
        logger.warning(reason)
        logger.warning(f"Dropping object {object_id} from the dataset")

        del self.files[object_id]
        del self.dims[object_id]
        self.prune_count += 1

    def _check_file_dimensions(self) -> tuple[int, int]:
        """Class initialization helper. Find the maximal pixel size that all images can support

        It is assumed that all the cutouts will be of very similar size; however, HSC's cutout
        server does not return exactly the same number of pixels for every query, even when it
        is given the same angular spread for every cutout.

        Machine learning models expect all images to be the same size.

        This function warns on significant differences (>2px) on any dimension between the largest
        and smallest images.

        Returns
        -------
        tuple(int,int)
            The minimum width and height in pixels of the entire dataset. In other words: the maximal image
            size in pixels that can be generated from ALL cutout images via cropping.
        """
        logger.info("Checking file dimensions to determine standard cutout size...")

        # Find the maximal cutout size that all images can support
        all_widths = [shape[0] for shape_list in self.dims.values() for shape in shape_list]
        all_heights = [shape[1] for shape_list in self.dims.values() for shape in shape_list]
        all_dimensions = all_widths + all_heights
        cutout_height = np.min(all_dimensions)
        cutout_width = cutout_height

        if (
            np.abs(cutout_width - np.mean(all_widths)) > 1
            or np.abs(cutout_height - np.mean(all_heights)) > 1
            or np.abs(np.max(all_widths) - np.mean(all_widths)) > 1
            or np.abs(np.max(all_heights) - np.mean(all_heights)) > 1
        ):
            msg = "Some images differ from the mean width or height of all images by more than 1px\n"
            msg += f"Images will be cropped to ({cutout_width}px, {cutout_height}px)\n"
            try:
                min_width_file = self._get_file(int(np.argmin(all_widths)))
                min_height_file = self._get_file(int(np.argmin(all_heights)))
                msg += f"See {min_width_file} for an example image of width {cutout_width}px\n"
                msg += f"See {min_height_file} for an example image of height {cutout_height}px"
            finally:
                logger.warning(msg)

        if min(cutout_height, cutout_width) < 1:
            msg = "Automatic determination found an absurd dimension of "
            msg += f"({cutout_width}px, {cutout_height}px)\n"
            msg += "Please either correct the data source or set a static cutout size with the \n"
            msg += "crop_to configuration in the [data_set] section of your hyrax config.\n"
            raise RuntimeError(msg)

        return cutout_width, cutout_height

    def _rebuild_manifest(self, config):
        from astropy.table import Table

        from hyrax.download import Downloader
        from hyrax.downloadCutout.downloadCutout import (
            parse_bool,
            parse_degree,
            parse_latitude,
            parse_longitude,
            parse_rerun,
            parse_tract_opt,
            parse_type,
        )

        if self.filter_catalog:
            raise RuntimeError("Cannot rebuild manifest. Set the filter_catalog=false and rerun")

        logger.info("Reading in catalog file... ")
        location_table = Downloader.filterfits(
            Path(config["download"]["fits_file"]).expanduser().resolve(), ["object_id", "ra", "dec"]
        )

        obj_to_ra = {
            str(location_table["object_id"][index]): location_table["ra"][index]
            for index in range(len(location_table))
        }
        obj_to_dec = {
            str(location_table["object_id"][index]): location_table["dec"][index]
            for index in range(len(location_table))
        }

        del location_table

        logger.info("Assembling Manifest...")

        # These are the column names expected in a manifest file by the downloader
        column_names = Downloader.MANIFEST_COLUMN_NAMES
        columns = {column_name: [] for column_name in column_names}

        # These will vary every object and must be implemented below
        dynamic_column_names = ["object_id", "filter", "dim", "tract", "ra", "dec", "filename"]
        # These are pulled from config ("sw", "sh", "rerun", "type", "image", "mask", and "variance")
        static_column_names = [name for name in column_names if name not in dynamic_column_names]

        # Check that all column names we need for a manifest are either in static or dynamic columns
        for column_name in column_names:
            if column_name not in static_column_names and column_name not in dynamic_column_names:
                raise RuntimeError(f"Error Assembling manifest {column_name} not implemented")

        static_values = {
            "sw": parse_degree(config["download"]["sw"]),
            "sh": parse_degree(config["download"]["sh"]),
            "rerun": parse_rerun(config["download"]["rerun"]),
            "type": parse_type(config["download"]["type"]),
            "image": parse_bool(config["download"]["image"]),
            "mask": parse_bool(config["download"]["mask"]),
            "variance": parse_bool(config["download"]["variance"]),
        }

        for index, (object_id, filter, filename, dim) in enumerate(self._all_files_full()):
            for static_col in static_column_names:
                columns[static_col].append(static_values[static_col])

            for dynamic_col in dynamic_column_names:
                if dynamic_col == "object_id":
                    columns[dynamic_col].append(int(object_id))
                elif dynamic_col == "filter":
                    columns[dynamic_col].append(filter)
                elif dynamic_col == "dim":
                    columns[dynamic_col].append(dim)
                elif dynamic_col == "tract":
                    # There's value in pulling tract from the filename rather than the download catalog
                    # in case The catalog had it wrong, the filename will have the value the cutout server
                    # provided.
                    tract = filename.split("_")[4]
                    columns[dynamic_col].append(parse_tract_opt(tract))
                elif dynamic_col == "ra":
                    ra = obj_to_ra[object_id]
                    columns[dynamic_col].append(parse_longitude(ra))
                elif dynamic_col == "dec":
                    dec = obj_to_dec[object_id]
                    columns[dynamic_col].append(parse_latitude(dec))
                elif dynamic_col == "filename":
                    columns[dynamic_col].append(filename)
                else:
                    # The tower of if statements has been entirely to create this failure path.
                    # which will be hit when someone alters dynamic column names above without also
                    # writing an implementation.
                    raise RuntimeError(f"No implementation to process column {dynamic_col}")
            if index != 0 and index % 1_000_000 == 0:
                logger.info(f"Addeed {index} objects to manifest")
        else:
            logger.info(f"Addeed {index + 1} objects to manifest")

        logger.info("Writing rebuilt manifest...")
        manifest_table = Table(columns)

        manifest_file_path = Path(config["general"]["data_dir"]) / Downloader.MANIFEST_FILE_NAME

        # Rename the old manifest
        if manifest_file_path.exists():
            filename_safe_now = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            new_file_name = Downloader.MANIFEST_FILE_NAME + f".archived.at.{filename_safe_now}"
            manifest_file_path.rename(Path(config["general"]["data_dir"]) / new_file_name)

        # Replace the old manifest
        manifest_table.write(manifest_file_path, overwrite=True, format="fits")

    def __contains__(self, object_id: str) -> bool:
        """Allows you to do `object_id in dataset` queries. Used by testing code.

        Parameters
        ----------
        object_id : str
            The object ID you'd like to know if is in the dataset

        Returns
        -------
        bool
            True of the object_id given is in the data set
        """
        return object_id in list(self.files.keys()) and object_id in list(self.dims.keys())

    def _all_files_full(self):
        """
        Private read-only iterator over all files that enforces a strict total order across
        objects and filters. Will not work prior to self.files, and self.path initialization in __init__

        Yields
        ------
        Tuple[object_id, filter, filename, dim]
            Members of this tuple are
            - The object_id as a string
            - The filter name as a string
            - The filename relative to self.path
            - A tuple containing the dimensions of the fits file in pixels.
        """
        for object_id in self.ids():
            dims = self.dims[object_id]
            for idx, (filter, filename) in enumerate(self._filter_filename(object_id)):
                yield (object_id, filter, filename, dims[idx])

    def _object_files(self, object_id):
        """
        Private read-only iterator over all files for a given object. This enforces a strict total order
        across filters. Will not work prior to self.files, and self.path initialization in __init__

        Guaranteed to only return files that have filters in self.filters_ref.

        Yields
        ------
        Path
            The path to the file.
        """
        for filter, filename in self._filter_filename(object_id):
            if filter in self.filters_ref:
                yield self._file_to_path(filename)



import json
import logging
import re
from collections import Counter
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.table import Table, join
from tqdm import tqdm

logger = logging.getLogger(__name__)


def save_umap_json(
    results_dir,
    output_json="umap_data.json",
    fits_table_path=None,
    columns_to_include=None,
    id_column="object_id",
    join_type="left",
    keep_first_match_only=False,
):
    """
    Saves UMAP 3D embeddings and object IDs into a JSON file for use in Three.js
    Also provides the option of including metadata from any Astropy readable table

    Parameters
    ----------
    results_dir : str or Path
        Directory containing the UMAP embedding batch files
    output_json : str, optional
        Output JSON filename, default is "umap_data.json"
    fits_table_path : str or Path, optional
        Path to a table in a file format that is supported by Astropy FITS table
        or pickle file (FITS table or Pandas dataframe) for cross-matching with
        additional properties
    columns_to_include : list, optional
        List of column names from the FITS table to include in the JSON output.
        If None and fits_table_path is provided, all columns will be included.
    id_column : str, optional
        Name of the column in the FITS table to use for ID matching, default is "object_id"
    join_type : str, optional
        Type of join to perform with the FITS table. Options are:
        - 'left': Include all UMAP points, with FITS data when available (default)
        - 'inner': Include only UMAP points that have matching FITS data
        - 'outer': Include all UMAP points and all FITS data, with nulls when no match
        - 'right': Include all FITS rows, with UMAP data when available
    keep_first_match_only : bool, optional
        If True, when multiple rows in the FITS table match a single object_id,
        only the first match will be kept. Default is False.
    """

    # Validate join_type parameter
    valid_join_types = ["left", "right", "inner", "outer"]
    if join_type not in valid_join_types:
        logger.error(f"join_type must be one of {valid_join_types}, got '{join_type}'")

    results_dir = Path(results_dir)

    # Find batch files matching 'batch_<number>.npy'
    batch_files = sorted([f for f in results_dir.glob("batch_*.npy") if re.match(r"batch_\d+\.npy$", f.name)])

    if not batch_files:
        raise FileNotFoundError(
            f"No valid batch files found in {results_dir}. Ensure the directory\
                                contains files matching batch_<number>.npy"
        )

    embeddings_list = []
    object_ids_list = []

    for batch_file in tqdm(batch_files, desc="Loading batch files"):
        data = np.load(batch_file)
        embeddings_list.append(data["tensor"])
        object_ids_list.append(data["id"])

    # Concatenate all embeddings and object IDs
    embeddings = np.concatenate(embeddings_list, axis=0)
    object_ids = np.concatenate(object_ids_list, axis=0)

    # Create an Astropy table with the embeddings and object IDs
    umap_table = Table()
    umap_table["id"] = object_ids
    umap_table["x"] = embeddings[:, 0]
    umap_table["y"] = embeddings[:, 1]
    umap_table["z"] = embeddings[:, 2]

    # Load FITS table if provided
    if fits_table_path is not None:
        fits_table_path = Path(fits_table_path)
        if not fits_table_path.exists():
            logger.error(f"Table file not found: {fits_table_path}")

        logger.info(f"Loading table file: {fits_table_path}")

        # Check if the file is a pickle file by examining its extension
        file_ext = fits_table_path.suffix.lower()

        if file_ext in [".pkl", ".pickle"]:
            import pickle

            # Load the pickle file
            logger.info(f"Detected pickle file format: {file_ext}")
            with open(fits_table_path, "rb") as f:
                # If it's a pandas DataFrame, convert to Astropy Table
                pickle_data = pickle.load(f)
                if isinstance(pickle_data, pd.DataFrame):
                    fits_table = Table.from_pandas(pickle_data)
                    logger.info("Converted pickled pandas DataFrame to Astropy Table")
                elif isinstance(pickle_data, Table):
                    fits_table = pickle_data
                    logger.info("Loaded pickled Astropy Table")
                else:
                    raise TypeError(
                        f"Pickled data must be either pandas DataFrame\
                                        or Astropy Table, got {type(pickle_data)}"
                    )
        else:
            # Try to load as a standard Astropy table format
            fits_table = Table.read(fits_table_path)

        # Verify the table has the specified ID column for cross-matching
        if id_column not in fits_table.colnames:
            raise ValueError(f"Table must contain a '{id_column}' column for cross-matching")

        # Check for duplicate IDs in the FITS table
        id_counts = Counter(fits_table[id_column])
        duplicate_ids = {id_val: count for id_val, count in id_counts.items() if count > 1}

        if duplicate_ids and keep_first_match_only:
            total_duplicates = sum(count - 1 for count in duplicate_ids.values())
            num_duplicated_ids = len(duplicate_ids)
            total_rows = len(fits_table)

            logger.info(
                f"Found {num_duplicated_ids} IDs with multiple rows ({total_duplicates} duplicate\
                    rows out of {total_rows} total rows)"
            )
            logger.info("Using keep_first_match_only=True, only the first occurrence of each ID will be kept")

            # Create a new table with only the first occurrence of each ID
            seen_ids = set()
            rows_to_keep = []

            for row in fits_table:
                current_id = row[id_column]
                if current_id not in seen_ids:
                    rows_to_keep.append(row)
                    seen_ids.add(current_id)

            # Create a new table with only the selected rows
            fits_table = Table(rows=rows_to_keep, names=fits_table.colnames)
            logger.info(f"Reduced FITS table from {total_rows} to {len(fits_table)} rows")
        elif duplicate_ids:
            total_duplicates = sum(count - 1 for count in duplicate_ids.values())
            num_duplicated_ids = len(duplicate_ids)

            logger.info(
                f"Warning: Found {num_duplicated_ids} IDs with multiple rows ({total_duplicates}\
                            duplicate rows)"
            )
            logger.info(
                "This will result in duplicate points in the output JSON. Use\
                            keep_first_match_only=True to keep only first matches."
            )

            # Example of duplicates for the user to understand
            if num_duplicated_ids > 0:
                example_id = list(duplicate_ids.keys())[0]
                example_count = duplicate_ids[example_id]
                logger.info(f"Example: ID {example_id} appears {example_count} times")

        # If columns_to_include is None, use all columns except the ID column
        if columns_to_include is None:
            columns_to_include = [col for col in fits_table.colnames if col != id_column]
            logger.info(f"Including all table columns: {columns_to_include}")
        else:
            # Verify requested columns exist in the table
            missing_columns = [col for col in columns_to_include if col not in fits_table.colnames]
            if missing_columns:
                raise ValueError(f"Requested columns not found in table: {missing_columns}")

        # Select only the columns we need from the FITS table
        columns_to_keep = [id_column] + columns_to_include
        fits_table = fits_table[columns_to_keep]

        # Rename the ID column in the FITS table to match the umap_table
        fits_table.rename_column(id_column, "id")

        # Check the data type of the FITS table ID column
        fits_id_dtype = fits_table["id"].dtype

        # Convert UMAP table ID column to match the FITS table type
        logger.info(f"Converting UMAP ID column to match FITS table ID type: {fits_id_dtype}")
        try:
            umap_table["id"] = umap_table["id"].astype(fits_id_dtype)
            logger.info("ID column conversion successful")
        except Exception as e:
            logger.info(f"Error converting ID column: {e}")
            logger.info("Falling back to string conversion for both tables")
            umap_table["id"] = umap_table["id"].astype(str)
            fits_table["id"] = fits_table["id"].astype(str)

        # Perform the join using Astropy's built-in functionality
        logger.info(f"Cross-matching using '{id_column}' column with join_type='{join_type}'...")
        joined_table = join(umap_table, fits_table, keys="id", join_type=join_type)

        # Report match statistics based on join type
        if join_type in ["left", "outer"]:
            total_umap_points = len(umap_table)
            missing_fits_data = sum(
                1
                for row in joined_table
                if any(
                    np.ma.is_masked(row[col]) for col in columns_to_include if col in joined_table.colnames
                )
            )
            matched_points = total_umap_points - missing_fits_data
            logger.info(
                f"UMAP points with matching table data: {matched_points} out of\
                         {total_umap_points} ({matched_points / total_umap_points * 100:.1f}%)"
            )

        if join_type in ["right", "outer"]:
            total_fits_rows = len(fits_table)
            missing_umap_data = sum(
                1
                for row in joined_table
                if any(np.ma.is_masked(row[col]) for col in ["x", "y", "z"] if col in joined_table.colnames)
            )
            matched_fits = total_fits_rows - missing_umap_data
            logger.info(
                f"Table rows with matching UMAP data: {matched_fits} out of\
                            {total_fits_rows} ({matched_fits / total_fits_rows * 100:.1f}%)"
            )

        if join_type == "inner":
            logger.info(f"Inner join results: {len(joined_table)} matching points")

        # Check if the join resulted in more rows than original UMAP points (meaning duplicates)
        if len(joined_table) > len(umap_table) and not keep_first_match_only:
            logger.info(
                f"Join resulted in {len(joined_table)} rows, which is more than the\
                         {len(umap_table)} UMAP points"
            )
            logger.info(
                "This indicates some UMAP points matched multiple table rows.\
                            The output will contain duplicate points."
            )

        # Convert to pandas DataFrame for easier handling of masked values
        df = joined_table.to_pandas()

        # Replace masked values with None for proper JSON serialization
        df = df.replace({pd.NA: None})

        # Convert numpy types to Python native types for JSON serialization
        for col in df.columns:
            if pd.api.types.is_numeric_dtype(df[col]):
                # Only convert to float for actual float columns, preserve ints
                if pd.api.types.is_integer_dtype(df[col]):
                    df[col] = df[col].astype("int64")  # or leave as is
                else:
                    df[col] = df[col].astype("float64")
    else:
        # If no FITS table provided, just convert UMAP data to pandas
        df = umap_table.to_pandas()

    # Convert to JSON format
    logger.info("Creating JSON data structure...")

    # First convert DataFrame to records style dict
    records = df.to_dict(orient="records")

    # Process each record to handle any remaining special types
    json_data = {"points": []}
    for record in tqdm(records, desc="Formatting JSON"):
        # Process any remaining numpy arrays or other special types
        for k, v in record.items():
            if isinstance(v, np.ndarray):
                record[k] = v.tolist()
            elif isinstance(v, (np.integer, np.floating)):
                record[k] = float(v)
            elif isinstance(v, bytes):
                try:
                    # Try to decode bytes to string
                    record[k] = v.decode("utf-8")
                except UnicodeDecodeError:
                    # If it can't be decoded, convert to base64 string
                    import base64

                    record[k] = base64.b64encode(v).decode("ascii")
            elif pd.isna(v):
                record[k] = None
        json_data["points"].append(record)

    # Save to file
    with open(output_json, "w") as f:
        json.dump(json_data, f, indent=2)

    logger.info(f"UMAP data saved to {output_json}")

    # Report overall statistics
    logger.info(f"Total points in JSON: {len(json_data['points'])}")
    if fits_table_path is not None:
        columns = ["x", "y", "z", "id"] + [c for c in columns_to_include if c in df.columns]
        logger.info(f"Each point has properties: {', '.join(columns)}")


if __name__ == "__main__":
    import argparse

    # Set up logging
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Create argument parser
    parser = argparse.ArgumentParser(description="Save UMAP 3D embeddings to JSON")

    # Required arguments
    parser.add_argument("results_dir", type=str, help="Directory containing the UMAP embedding batch files")

    # Optional arguments
    parser.add_argument(
        "--output_json",
        "-o",
        type=str,
        default="umap_data.json",
        help="Output JSON filename (default: umap_data.json)",
    )
    parser.add_argument(
        "--fits_table_path",
        "-f",
        type=str,
        default=None,
        help="Path to a table format accepted by Astropy Tables or pickle file for\
                            cross-matching and incorproation of additional metadata",
    )
    parser.add_argument(
        "--columns_to_include",
        "-c",
        type=str,
        nargs="+",
        default=None,
        help="List of column names from the fits-table to include",
    )
    parser.add_argument(
        "--id_column",
        "-i",
        type=str,
        default="object_id",
        help="Name of the ID column for matching to the fits-table (default: object_id)",
    )
    parser.add_argument(
        "--join_type",
        "-j",
        type=str,
        default="left",
        choices=["left", "right", "inner", "outer"],
        help="Type of join to perform (default: left)",
    )
    parser.add_argument(
        "--keep_first_match_only",
        "-k",
        action="store_true",
        help="Keep only the first match when multiple rows match an ID in the fits table",
    )

    args = parser.parse_args()

    # Run the function with the provided arguments
    save_umap_json(
        results_dir=args.results_dir,
        output_json=args.output_json,
        fits_table_path=args.fits_table_path,
        columns_to_include=args.columns_to_include,
        id_column=args.id_column,
        join_type=args.join_type,
        keep_first_match_only=args.keep_first_match_only,
    )



import logging

import torch.nn as nn

from .model_registry import hyrax_model

logger = logging.getLogger()


@hyrax_model
class HyraxLoopback(nn.Module):
    """Simple model for testing which returns its own input"""

    def __init__(self, config, shape):
        from functools import partial

        super().__init__()
        # The optimizer needs at least one weight, so we add a dummy module here
        self.unused_module = nn.Linear(1, 1)
        self.config = config

        def load(self, weight_file):
            """Boilerplate function to load weights. However, this model has no
            weights so we do nothing."""
            pass

        # We override this way rather than defining a method because
        # Torch has some __init__ related cleverness which stomps our
        # load definition when performed in the usual fashion.
        self.load = partial(load, self)

    def forward(self, x):
        """We simply return our input"""
        if isinstance(x, tuple):
            # if x is a tuple, extract the first element (it should be a tensor)
            x, _ = x
        return x

    def train_step(self, batch):
        """Training is a noop"""
        logger.debug(f"Batch length: {len(batch)}")
        return {"loss": 0.0}



import gc
import logging
import os
import pickle
import warnings
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional, Union

import numpy as np
import psutil

from .verb_registry import Verb, hyrax_verb

logger = logging.getLogger(__name__)


@hyrax_verb
class Umap(Verb):
    """Umap latent space points into 2d"""

    cli_name = "umap"
    add_parser_kwargs = {}

    @staticmethod
    def setup_parser(parser: ArgumentParser):
        """Stub of parser setup"""
        parser.add_argument(
            "-i",
            "--input-dir",
            type=str,
            required=False,
            help="Directory containing inference results to umap.",
        )

    # Should there be a version of this on the base class which uses a dict on the Verb
    # superclass to build the call to run based on what the subclass verb defined in setup_parser
    def run_cli(self, args: Optional[Namespace] = None):
        """Stub CLI implementation"""
        logger.info("umap run from cli")
        if args is None:
            raise RuntimeError("Run CLI called with no arguments.")

        # This is where we map from CLI parsed args to a
        # self.run (args) call.
        return self.run(input_dir=args.input_dir)

    def run(self, input_dir: Optional[Union[Path, str]] = None):
        """
        Create a umap of a particular inference run

        This method loads the latent space representations from an inference run,
        samples a subset of data points, flattens them if necessary, and then fits
        a UMAP model. The fitted reducer is then used to transform the entire dataset
        into a lower-dimensional space.

        Parameters
        ----------
        input_dir : str or Path, Optional
            The directory containing the inference results.

        Returns
        -------
        None
            The method does not return anything but saves the UMAP representations to disk.
        """
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            return self._run(input_dir)

    def _run(self, input_dir: Optional[Union[Path, str]] = None):
        """See run()"""
        import multiprocessing as mp

        import umap
        from tqdm.auto import tqdm

        from hyrax.config_utils import create_results_dir
        from hyrax.data_sets.inference_dataset import InferenceDataSet, InferenceDataSetWriter

        self.reducer = umap.UMAP(**self.config["umap.UMAP"])

        # Load all the latent space data.
        inference_results = InferenceDataSet(self.config, results_dir=input_dir)
        total_length = len(inference_results)

        # Set up the results directory where we will store our umapped output
        results_dir = create_results_dir(self.config, "umap")
        logger.info(f"Saving UMAP results to {results_dir}")
        umap_results = InferenceDataSetWriter(inference_results, results_dir)

        # Sample the data to fit
        config_sample_size = self.config["umap"]["fit_sample_size"]
        sample_size = int(np.min([config_sample_size if config_sample_size else np.inf, total_length]))
        rng = np.random.default_rng()
        index_choices = rng.choice(np.arange(total_length), size=sample_size, replace=False)

        # If the input to umap is not of the shape [samples,input_dims] we reshape the input accordingly
        data_sample = inference_results[index_choices].numpy().reshape((sample_size, -1))

        self._log_memory_usage("Before fitting umap")
        logger.info("Fitting the UMAP")
        # Fit a single reducer on the sampled data
        self.reducer.fit(data_sample)
        self._log_memory_usage("After fitting umap")

        # Save the reducer to our results directory
        if self.config["umap"]["save_fit_umap"]:
            logger.info("Saving fitted UMAP Reducer")
            with open(results_dir / "umap.pickle", "wb") as f:
                pickle.dump(self.reducer, f)

        # Reclaim Memory
        del data_sample
        gc.collect()
        self._log_memory_usage("After Garbage Collection")

        # Run all data through the reducer in batches, writing it out as we go.
        batch_size = self.config["data_loader"]["batch_size"]
        num_batches = int(np.ceil(total_length / batch_size))

        all_indexes = np.arange(0, total_length)
        all_ids = np.array(list(inference_results.ids()))

        # Generator expression that gives a batch tuple composed of:
        # batch ids, inference results
        args = (
            (
                all_ids[batch_indexes],
                # We flatten all dimensions of the input array except the dimension
                # corresponding to batch elements. This ensures that all inputs to
                # the UMAP algorithm are flattend per input item in the batch
                inference_results[batch_indexes].numpy().reshape(len(batch_indexes), -1),
            )
            for batch_indexes in np.array_split(all_indexes, num_batches)
        )

        if self.config["umap"]["parallel"]:
            # Process pool loop
            # Use 'spawn' context to safely create subprocesses after
            # OpenMP threads are being opened by other processes in hyrax
            # Not using spawn causes the issue linked below
            # https://github.com/lincc-frameworks/hyrax/issues/291
            # TODO: Find more elegant solution than just using spawn
            with mp.get_context("spawn").Pool(processes=mp.cpu_count()) as pool:
                # iterate over the mapped results to write out the umapped points
                # imap returns results as they complete so writing should complete
                # in parallel for large datasets
                for batch_ids, transformed_batch in tqdm(
                    pool.imap(self._transform_batch, args),
                    desc="Creating lower dimensional representation using UMAP:",
                    total=num_batches,
                ):
                    umap_results.write_batch(batch_ids, transformed_batch)
        else:
            # Sequential loop
            for batch_ids, batch in tqdm(
                args,
                desc="Creating lower dimensional representation using UMAP:",
                total=num_batches,
            ):
                transformed_batch = self.reducer.transform(batch)
                self._log_memory_usage(f"During transformation of batch of shape {batch.shape}")
                umap_results.write_batch(batch_ids, transformed_batch)

        umap_results.write_index()
        logger.info("Finished transforming all data through UMAP")

        return InferenceDataSet(self.config, results_dir)

    def _transform_batch(self, batch_tuple: tuple):
        """Private helper to transform a single batch

        Parameters
        ----------
        batch_tuple : tuple()
            first element is the IDs of the batch as a numpy array
            second element is the inference results to transform as a numpy array with shape (batch_len, N)
            where N is the total number of dimensions in the inference result. Caller flattens all inference
            result axes for us.

        Returns
        -------
        tuple
            first element is the ids of the batch as a numpy array
            second element is the results of running the umap transform on the input as a numpy array.
        """
        batch_ids, batch = batch_tuple
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            logger.debug("Transforming a batch ...")
            return (batch_ids, self.reducer.transform(batch))

    @staticmethod
    def _log_memory_usage(message: str = ""):
        """
        Log the current resident set size (RSS) memory usage of the current process in gigabytes.

        Parameters
        ----------
        message : str, optional
            A descriptive message to include in the log output for context.

        Notes
        -----
        This method is intended for debugging and performance monitoring.
        """
        process = psutil.Process(os.getpid())
        mem_gb = process.memory_info().rss / 1024**3
        logger.debug(f"{message} | Memory usage: {mem_gb:.2f} GB")



import functools
import logging
import warnings
from pathlib import Path
from typing import Any, Callable, Optional, Union

import ignite.distributed as idist
import numpy as np

with warnings.catch_warnings():
    warnings.simplefilter(action="ignore", category=DeprecationWarning)
    import mlflow

from collections.abc import Iterator, Sequence

import torch
from ignite.engine import Engine, EventEnum, Events
from ignite.handlers import Checkpoint, DiskSaver, global_step_from_engine
from ignite.handlers.tqdm_logger import ProgressBar
from tensorboardX import SummaryWriter
from torch.nn.parallel import DataParallel, DistributedDataParallel
from torch.utils.data import DataLoader, Dataset, Sampler

from hyrax.config_utils import ConfigDict
from hyrax.data_sets.data_set_registry import HyraxDataset, fetch_data_set_class
from hyrax.models.model_registry import fetch_model_class

logger = logging.getLogger(__name__)


class SubsetSequentialSampler(Sampler[int]):
    r"""Samples elements sequentially from a given list of indices, without replacement.

    Args:
        indices : sequence
            a sequence of indices
    """

    indices: Sequence[int]

    def __init__(self, indices: Sequence[int], generator=None) -> None:
        self.indices = indices
        self.generator = generator

    def __iter__(self) -> Iterator[int]:
        for i in self.indices:
            yield i

    def __len__(self) -> int:
        return len(self.indices)


def setup_dataset(config: ConfigDict, tensorboardx_logger: Optional[SummaryWriter] = None) -> Dataset:
    """Create a dataset object based on the configuration.

    Parameters
    ----------
    config : ConfigDict
        The entire runtime configuration
    tensorboardx_logger : SummaryWriter, optional
        If Tensorboard is in use, the tensorboard logger so the dataset can log things

    Returns
    -------
    Dataset
        An instance of the dataset class specified in the configuration
    """

    # Fetch data loader class specified in config and create an instance of it
    data_set_cls = fetch_data_set_class(config)
    data_set: HyraxDataset = data_set_cls(config)  # type: ignore[call-arg]

    data_set.tensorboardx_logger = tensorboardx_logger

    return data_set


def setup_model(config: ConfigDict, dataset: Dataset) -> torch.nn.Module:
    """Create a model object based on the configuration.

    Parameters
    ----------
    config : ConfigDict
        The entire runtime configuration
    dataset : Dataset
        Only used to determine the input shape of the data

    Returns
    -------
    torch.nn.Module
        An instance of the model class specified in the configuration
    """

    # Fetch model class specified in config and create an instance of it
    model_cls = fetch_model_class(config)
    model = model_cls(config=config, dataset=dataset)  # type: ignore[attr-defined]

    return model


def dist_data_loader(
    data_set: Dataset,
    config: ConfigDict,
    split: Union[str, list[str], bool] = False,
):
    """Create Pytorch Ignite distributed data loaders

    It is recommended that each verb needing dataloaders only call this function once.

    Parameters
    ----------
    data_set : Dataset
        A Pytorch Dataset object
    config : ConfigDict
        Hyrax runtime configuration
    split : Union[str, list[str]], Optional
        The name(s) of the split we want to use from the data set.
        If this is false or not passed, then a single data loader is returned
        that corresponds to the entire dataset.

    Returns
    -------
    Dataloader (or an ignite-wrapped equivalent)
        This is the distributed dataloader, formed by calling ignite.distributed.auto_dataloader

    For multiple splits, we return a dictionary where the keys are the names of the splits
    and the value is either a Dataloader as described above or the value None if the split
    was not configured.

    If an iterable dataset is passed, we cannot create multiple splits with a pyTorch sampler object
    so we return the same thing for all splits, which is a dataloader representing the entire iterable
    """
    # Handle case where no split is needed.
    if isinstance(split, bool):
        # We still need to return the list of indexes used by the dataloader,
        # but here, it will simply be the indexes for the entire dataset.
        if data_set.is_iterable():
            ids = list(data_set.ids())
            indexes = list(range(len(ids)))
        else:
            indexes = list(range(len(data_set)))

        # Note that when sampler=None, a default sampler is used. The default config
        # defines shuffle=False, which should prevent any shuffling of of the data.
        # We expect that this will be the primary use case when running inference.
        return idist.auto_dataloader(data_set, sampler=None, **config["data_loader"]), indexes

    # Sanitize split argument
    if isinstance(split, str):
        split = [split]

    # Configure the torch rng
    torch_rng = torch.Generator()
    seed = config["data_set"]["seed"] if config["data_set"]["seed"] else None
    if seed is not None:
        torch_rng.manual_seed(seed)

    if data_set.is_iterable():
        ids = list(data_set.ids())
        indexes = list(range(len(ids)))
        dataloaders = {
            s: (idist.auto_dataloader(data_set, pin_memory=True, **config["data_loader"]), indexes)
            for s in split
        }
    else:
        # Create the indexes for all splits based on config.
        indexes = create_splits(data_set, config)

        # Create samplers and dataloaders for each split we are interested in
        samplers = {s: SubsetSequentialSampler(indexes[s]) if indexes.get(s) else None for s in split}

        dataloaders = {
            split: (idist.auto_dataloader(data_set, sampler=sampler, **config["data_loader"]), indexes[split])
            if sampler
            else None
            for split, sampler in samplers.items()
        }

        none_keys = [k for k, v in dataloaders.items() if v is None]
        for key in none_keys:
            del dataloaders[key]

    # Return only one if we were only passed one split in, return the dictionary otherwise.
    return dataloaders[split[0]] if len(split) == 1 else dataloaders


def create_splits(data_set: Dataset, config: ConfigDict):
    """Returns train, test, and validation indexes constructed to be used with the passed in
    dataset. The allocation of indexes in the underlying dataset to samplers depends on
    the data_set section of the config dict.

    Parameters
    ----------
    data_set : Dataset
        The data set to use
    config : ConfigDict
        Configuration that defines dataset splits
    split : str
        Name of the split to use.
    """
    data_set_size = len(data_set)  # type: ignore[arg-type]

    # Init the splits based on config values
    train_size = config["data_set"]["train_size"] if config["data_set"]["train_size"] else None
    test_size = config["data_set"]["test_size"] if config["data_set"]["test_size"] else None
    validate_size = config["data_set"]["validate_size"] if config["data_set"]["validate_size"] else None

    # Convert all values specified as counts into ratios of the underlying container
    if isinstance(train_size, int):
        train_size = train_size / data_set_size
    if isinstance(test_size, int):
        test_size = test_size / data_set_size
    if isinstance(validate_size, int):
        validate_size = validate_size / data_set_size

    # Initialize Test size when not provided
    if test_size is None:
        if train_size is None:
            train_size = 0.25

        if validate_size is None:  # noqa: SIM108
            test_size = 1.0 - train_size
        else:
            test_size = 1.0 - (train_size + validate_size)

    # Initialize train size when not provided, and can be inferred from test_size and validate_size.
    if train_size is None:
        if validate_size is None:  # noqa: SIM108
            train_size = 1.0 - test_size
        else:
            train_size = 1.0 - (test_size + validate_size)

    # If splits cover more than the entire dataset, error out.
    if validate_size is None:
        if np.round(train_size + test_size, decimals=5) > 1.0:
            raise RuntimeError("Split fractions add up to more than 1.0")
    elif np.round(train_size + test_size + validate_size, decimals=5) > 1.0:
        raise RuntimeError("Split fractions add up to more than 1.0")

    # If any split is less than 0.0 also error out
    if (
        np.round(test_size, decimals=5) < 0.0
        or np.round(train_size, decimals=5) < 0.0
        or (validate_size is not None and np.round(validate_size, decimals=5) < 0.0)
    ):
        raise RuntimeError("One of the Split fractions configured is negative.")

    indices = list(range(data_set_size))

    # shuffle the indices
    seed = config["data_set"]["seed"] if config["data_set"]["seed"] else None
    np.random.seed(seed)
    np.random.shuffle(indices)

    # Given the number of samples in the dataset and the ratios of the splits
    # we can calculate the number of samples in each split.
    num_test = int(np.round(data_set_size * test_size))
    num_train = int(np.round(data_set_size * train_size))

    # split the indices
    test_idx = indices[:num_test]
    train_idx = indices[num_test : num_test + num_train]

    # assume that validate gets all the remaining indices
    if validate_size:
        num_validate = int(np.round(data_set_size * validate_size))
        valid_idx = indices[num_test + num_train : num_test + num_train + num_validate]

    split_inds = {"train": train_idx, "test": test_idx}
    if validate_size:
        split_inds["validate"] = valid_idx

    return split_inds


@functools.singledispatch
def _handle_nans(batch, config):
    """The default _handle_nan function. Will print a warning and return `batch`."""
    logger.warning(
        f"Encountered an unhandled batch type, {type(batch)}, while\
                   attempting to handle NaN values in the data."
    )
    return batch


@_handle_nans.register(torch.Tensor)
def _handle_nans_tensor(batch, config):
    """The implementation of _handle_nans when expecting `batch` to be a tensor."""
    return _handle_nans_logic(batch, config)


@_handle_nans.register(tuple)
def _handle_nans_tuple(batch, config):
    """This is the tuple-specific implementation of _handle_nans. The first element
    of the tuple is expected to hold the data that needs nan-cleaning."""
    nan_handled = _handle_nans_logic(batch[0], config)
    return (nan_handled, batch[1])


def _handle_nans_logic(batch, config):
    from torch import any, isnan

    if config["data_set"]["nan_mode"] is False:
        if any(isnan(batch)):
            msg = "Input data contains NaN values. This may mean your model output is all NaNs."
            msg += "Consider setting config['data_set']['nan_mode'] = 'quantile' or 'zero' or writing a "
            msg += "to_tensor() function for your model. Search hyrax readthedocs for 'to_tensor' "
            msg += "to get started."
            logger.warning(msg)
        return batch

    if config["data_set"]["nan_mode"] == "quantile":
        quantile = config["data_set"]["nan_quantile"]
        if quantile < 0.0 or quantile > 1.0:
            raise RuntimeError('set config["data_set"]["nan_quantile"] to a value between 0 and 1')
        return _handle_nan_quantile(batch, quantile)
    elif config["data_set"]["nan_mode"] == "zero":
        return _handle_nan_zero(batch)
    else:
        msg = f"nan mode was set to '{config['data_set']['nan_mode']}' which is unsupported."
        msg += "The supported modes are 'quantile' and 'zero'."
        raise NotImplementedError(msg)


def _handle_nan_quantile(batch, quantile):
    from torch import any, isnan

    if any(isnan(batch)):
        flat_batch = torch.reshape(batch, (batch.shape[0], -1))
        batch_quantile = torch.nanquantile(flat_batch, q=quantile, dim=-1)
        for i, val in enumerate(batch_quantile):
            batch[i] = torch.nan_to_num(batch[i], val)

    return batch


def _handle_nan_zero(batch):
    from torch import any, isnan

    if any(isnan(batch)):
        batch = torch.nan_to_num(batch, nan=0.0)

    return batch


def _inner_loop(func, to_tensor, device, config, engine, batch):
    """This wraps a model-specific function (func) to move data to the appropriate device."""
    # If we have a dict of lists, we need to decode the dict
    # This will give us either a tensor of the whole batch or a tuple
    if isinstance(batch, dict):
        batch = to_tensor(batch)

    batch = _handle_nans(batch, config)

    # Send the batch to the device
    batch = batch.to(device) if isinstance(batch, torch.Tensor) else tuple(i.to(device) for i in batch)
    return func(batch)


def _create_process_func(funcname, device, model, config):
    inner_step = extract_model_method(model, funcname)
    to_tensor = extract_model_method(model, "to_tensor")
    inner_loop = functools.partial(_inner_loop, inner_step, to_tensor, device, config)
    return inner_loop


def create_engine(funcname: str, device: torch.device, model: torch.nn.Module, config: ConfigDict) -> Engine:
    """Unified creation of the pytorch engine object for either an evaluator or trainer.

    This function will automatically unwrap a distributed model to find the necessary function, and construct
    the necessary functions to transfer data to the device on every batch, so model code can be the same no
    matter where the model is being run.

    Parameters
    ----------
    funcname : str
        The function name on the model that we will call in the core of the engine loop, and be called once
        per batch
    device : torch.device
        The device the engine will run the model on
    model : torch.nn.Module
        The Model the engine will be using
    config : ConfigDict
        The runtime config in use
    """
    return Engine(_create_process_func(funcname, device, model, config))


def extract_model_method(model, method_name):
    """Extract a method from a model, which may be wrapped in a DistributedDataParallel
    or DataParallel object. For instance, method_name could be `train_step` or
    `forward`.

    Parameters
    ----------
    model : nn.Module, DistributedDataParallel, or DataParallel
        The model to extract the method from
    method_name : str
        Name of the method to extract

    Returns
    -------
    Callable
        The method extracted from the model
    """
    wrapped = type(model) is DistributedDataParallel or type(model) is DataParallel
    return getattr(model.module if wrapped else model, method_name)


def create_evaluator(
    model: torch.nn.Module, save_function: Callable[[torch.Tensor, torch.Tensor], Any], config: ConfigDict
) -> Engine:
    """Creates an evaluator engine
    Primary purpose of this function is to attach the appropriate handlers to an evaluator engine

    Parameters
    ----------
    model : torch.nn.Module
        The model to evaluate

    save_function : Callable[[torch.Tensor], Any]
        A function which will receive Engine.state.output at the end of each iteration. The intent
        is for the results of evaluation to be saved.

    config : ConfigDict
        The runtime config in use

    Returns
    -------
    pytorch-ignite.Engine
        Engine object which when run will evaluate the model.
    """
    device = idist.device()
    model.eval()
    model = idist.auto_model(model)
    evaluator = create_engine("forward", device, model, config)

    @evaluator.on(Events.STARTED)
    def log_eval_start(evaluator):
        logger.info(f"Evaluating model on device: {device}")
        logger.info(f"Total epochs: {evaluator.state.max_epochs}")

    @evaluator.on(Events.ITERATION_COMPLETED)
    def log_iteration_complete(evaluator):
        save_function(evaluator.state.batch, evaluator.state.output)

    @evaluator.on(Events.COMPLETED)
    def log_total_time(evaluator):
        logger.info(f"Total evaluation time: {evaluator.state.times['COMPLETED']:.2f}[s]")

    pbar = ProgressBar(persist=False, bar_format="")
    pbar.attach(evaluator)

    return evaluator


#! There will likely be a significant amount of code duplication between the
#! `create_trainer` and `create_validator` functions. We should find a way to
#! refactor this code to reduce duplication.
def create_validator(
    model: torch.nn.Module,
    config: ConfigDict,
    results_directory: Path,
    tensorboardx_logger: SummaryWriter,
    validation_data_loader: DataLoader,
    trainer: Engine,
) -> Engine:
    """This function creates a Pytorch Ignite engine object that will be used to
    validate the model.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train
    config : ConfigDict
        Hyrax runtime configuration
    results_directory : Path
        The directory where training results will be saved
    tensorboardx_logger : SummaryWriter
        The tensorboard logger object
    validation_data_loader : DataLoader
        The data loader for the validation data
    trainer : Engine
        The engine object that will be used to train the model. We will use specific
        hooks in the trainer to determine when to run the validation engine.

    Returns
    -------
    pytorch-ignite.Engine
        Engine object that will be used to train the model.
    """

    device = idist.device()
    model = idist.auto_model(model)

    validator = create_engine("train_step", device, model, config)
    fixup_engine(validator)

    @validator.on(Events.STARTED)
    def set_model_to_eval_mode():
        model.eval()

    @validator.on(Events.COMPLETED)
    def set_model_to_train_mode():
        model.train()

    @validator.on(HyraxEvents.HYRAX_EPOCH_COMPLETED)
    def log_training_loss():
        logger.debug(f"Validation run time: {validator.state.times['EPOCH_COMPLETED']:.2f}[s]")
        logger.debug(f"Validation metrics: {validator.state.output}")

    @trainer.on(HyraxEvents.HYRAX_EPOCH_COMPLETED)
    def run_validation():
        validator.run(validation_data_loader)

    def log_validation_loss(validator, trainer):
        step = trainer.state.get_event_attrib_value(Events.EPOCH_COMPLETED)
        tensorboardx_logger.add_scalar("training/validation/loss", validator.state.output["loss"], step)
        mlflow.log_metrics({"validation/loss": validator.state.output["loss"]}, step=step)

    validator.add_event_handler(HyraxEvents.HYRAX_EPOCH_COMPLETED, log_validation_loss, trainer)

    return validator


def create_trainer(
    model: torch.nn.Module, config: ConfigDict, results_directory: Path, tensorboardx_logger: SummaryWriter
) -> Engine:
    """This function is originally copied from here:
    https://github.com/pytorch-ignite/examples/blob/main/tutorials/intermediate/cifar10-distributed.py#L164

    It was substantially trimmed down to make it easier to understand.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train
    config : ConfigDict
        Hyrax runtime configuration
    results_directory : Path
        The directory where training results will be saved
    tensorboardx_logger : SummaryWriter
        The tensorboard logger object

    Returns
    -------
    pytorch-ignite.Engine
        Engine object that will be used to train the model.
    """
    device = idist.device()
    model.train()
    model = idist.auto_model(model)
    trainer = create_engine("train_step", device, model, config)
    fixup_engine(trainer)

    optimizer = extract_model_method(model, "optimizer")

    to_save = {
        "model": model,
        "optimizer": optimizer,
        "trainer": trainer,
    }

    #! We may want to move the checkpointing logic over to the `validator`.
    #! It was created here initially because this was the only place where the
    #! model training was happening.
    latest_checkpoint = Checkpoint(
        to_save,
        DiskSaver(results_directory, require_empty=False),
        n_saved=1,
        global_step_transform=global_step_from_engine(trainer, Events.EPOCH_COMPLETED),
        filename_pattern="{name}_epoch_{global_step}.{ext}",
    )

    def neg_loss_score(engine):
        return -engine.state.output["loss"]

    best_checkpoint = Checkpoint(
        to_save,
        DiskSaver(results_directory, require_empty=False),
        n_saved=1,
        global_step_transform=global_step_from_engine(trainer, Events.EPOCH_COMPLETED),
        score_name="loss",
        score_function=neg_loss_score,
        greater_or_equal=True,
    )

    if config["train"]["resume"]:
        prev_checkpoint = torch.load(config["train"]["resume"], map_location=device)
        Checkpoint.load_objects(to_load=to_save, checkpoint=prev_checkpoint)

    # results_root_dir = Path(config["general"]["results_dir"]).expanduser().resolve()
    # mlflow_logger = MLflowLogger("file://" + str(results_root_dir / "mlflow"))

    @trainer.on(Events.STARTED)
    def log_training_start(trainer):
        logger.info(f"Training model on device: {device}")

    @trainer.on(Events.EPOCH_STARTED)
    def log_epoch_start(trainer):
        logger.debug(f"Starting epoch {trainer.state.epoch}")

    @trainer.on(Events.ITERATION_COMPLETED(every=10))
    def log_training_loss_tensorboard(trainer):
        step = trainer.state.get_event_attrib_value(Events.ITERATION_COMPLETED)
        tensorboardx_logger.add_scalar("training/training/loss", trainer.state.output["loss"], step)
        mlflow.log_metrics({"training/loss": trainer.state.output["loss"]}, step=step)

    @trainer.on(HyraxEvents.HYRAX_EPOCH_COMPLETED)
    def log_training_loss(trainer):
        logger.debug(f"Epoch {trainer.state.epoch} run time: {trainer.state.times['EPOCH_COMPLETED']:.2f}[s]")
        logger.debug(f"Epoch {trainer.state.epoch} metrics: {trainer.state.output}")

    trainer.add_event_handler(HyraxEvents.HYRAX_EPOCH_COMPLETED, latest_checkpoint)
    trainer.add_event_handler(HyraxEvents.HYRAX_EPOCH_COMPLETED, best_checkpoint)

    @trainer.on(Events.COMPLETED)
    def log_total_time(trainer):
        logger.info(f"Total training time: {trainer.state.times['COMPLETED']:.2f}[s]")

    def log_last_checkpoint_location(_, latest_checkpoint):
        logger.info(f"Latest checkpoint saved as: {latest_checkpoint.last_checkpoint}")

    def log_best_checkpoint_location(_, best_checkpoint):
        logger.info(f"Best metric checkpoint saved as: {best_checkpoint.last_checkpoint}")

    trainer.add_event_handler(Events.COMPLETED, log_last_checkpoint_location, latest_checkpoint)
    trainer.add_event_handler(Events.COMPLETED, log_best_checkpoint_location, best_checkpoint)

    pbar = ProgressBar(persist=False, bar_format="")
    pbar.attach(trainer)

    return trainer


class HyraxEvents(EventEnum):
    """
    Workaround event for a pytorch ignite bug. See fixup_engine for details
    """

    HYRAX_EPOCH_COMPLETED = "HyraxEpochCompleted"


def fixup_engine(engine: Engine) -> Engine:
    """
    Workaround for this pytorch ignite bug (https://github.com/pytorch/ignite/issues/3372) where
    engine.state.output is not available at EPOCH_COMPLETED or later times (COMPLETED, etc)

    We create a new event HYRAX_EPOCH_COMPLETED which triggers at ITERATION_COMPLETED, but only on the final
    iteration. This is just before the erronious state reset.

    This hack relies on pytorch ignite internal state, but can be removed as soon as our fix is mainlined
    (https://github.com/pytorch/ignite/pull/3373) in version 0.6.0 estimated August 2025
    """
    from more_itertools import peekable

    engine.register_events(*HyraxEvents)

    @engine.on(Events.ITERATION_COMPLETED)
    def maintain_event_handler(engine):
        # Ensure we have a peekable iterator in the engine.
        if not hasattr(engine._dataloader_iter, "peek"):
            # Replace with a pass-through peekable iterator
            engine._dataloader_iter = peekable(engine._dataloader_iter)

        # On the last iteration the peekable iterator evaluates as true
        if not engine._dataloader_iter:
            engine.fire_event(HyraxEvents.HYRAX_EPOCH_COMPLETED)
